# ruff: noqa: UP006 UP007
import contextlib
import dataclasses as dc
import enum
import faulthandler
import gc
import importlib
import logging
import os
import pwd
import resource
import signal
import sys
import typing as ta

from .. import lang
from .base import Bootstrap
from .base import ContextBootstrap
from .base import SimpleBootstrap


if ta.TYPE_CHECKING:
    from .. import libc
    from .. import logs
    from ..formats import dotenv
    from ..lite import pidfile

else:
    libc = lang.proxy_import('..libc', __package__)
    logs = lang.proxy_import('..logs', __package__)
    dotenv = lang.proxy_import('..formats.dotenv', __package__)
    pidfile = lang.proxy_import('..lite.pidfile', __package__)


##


class CwdBootstrap(ContextBootstrap['CwdBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        path: ta.Optional[str] = None

    @contextlib.contextmanager
    def enter(self) -> ta.Iterator[None]:
        if self._config.path is not None:
            prev = os.getcwd()
            os.chdir(self._config.path)
        else:
            prev = None

        try:
            yield

        finally:
            if prev is not None:
                os.chdir(prev)


##


class SetuidBootstrap(SimpleBootstrap['SetuidBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        user: ta.Optional[str] = None

    def run(self) -> None:
        if self._config.user is not None:
            user = pwd.getpwnam(self._config.user)
            os.setuid(user.pw_uid)


##


class GcDebugFlag(enum.Enum):
    STATS = gc.DEBUG_STATS
    COLLECTABLE = gc.DEBUG_COLLECTABLE
    UNCOLLECTABLE = gc.DEBUG_UNCOLLECTABLE
    SAVEALL = gc.DEBUG_SAVEALL
    LEAK = gc.DEBUG_LEAK


class GcBootstrap(ContextBootstrap['GcBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        disable: bool = False
        debug: ta.Optional[int] = None

    @contextlib.contextmanager
    def enter(self) -> ta.Iterator[None]:
        prev_enabled = gc.isenabled()
        if self._config.disable:
            gc.disable()

        if self._config.debug is not None:
            prev_debug = gc.get_debug()
            gc.set_debug(self._config.debug)
        else:
            prev_debug = None

        try:
            yield

        finally:
            if prev_enabled:
                gc.enable()

            if prev_debug is not None:
                gc.set_debug(prev_debug)


##


class NiceBootstrap(SimpleBootstrap['NiceBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        nice: ta.Optional[int] = None

    def run(self) -> None:
        if self._config.nice is not None:
            os.nice(self._config.nice)


##


class LogBootstrap(ContextBootstrap['LogBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        level: ta.Union[str, int, None] = None
        json: bool = False

    @contextlib.contextmanager
    def enter(self) -> ta.Iterator[None]:
        if self._config.level is None:
            yield
            return

        handler = logs.configure_standard_logging(
            self._config.level,
            json=self._config.json,
        )

        try:
            yield

        finally:
            if handler is not None:
                logging.root.removeHandler(handler)


##


class FaulthandlerBootstrap(ContextBootstrap['FaulthandlerBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        enabled: ta.Optional[bool] = None

    @contextlib.contextmanager
    def enter(self) -> ta.Iterator[None]:
        if self._config.enabled is None:
            yield
            return

        prev = faulthandler.is_enabled()
        if self._config.enabled:
            faulthandler.enable()
        else:
            faulthandler.disable()

        try:
            yield

        finally:
            if prev:
                faulthandler.enable()
            else:
                faulthandler.disable()


##


SIGNALS_BY_NAME = {
    a[len('SIG'):]: v  # noqa
    for a in dir(signal)
    if a.startswith('SIG')
    and not a.startswith('SIG_')
    and a == a.upper()
    and isinstance((v := getattr(signal, a)), int)
}


class PrctlBootstrap(SimpleBootstrap['PrctlBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        dumpable: bool = False
        deathsig: ta.Union[int, str, None] = None

    def run(self) -> None:
        if self._config.dumpable:
            libc.prctl(libc.PR_SET_DUMPABLE, 1, 0, 0, 0, 0)

        if self._config.deathsig is not None:
            if isinstance(self._config.deathsig, int):
                sig = self._config.deathsig
            else:
                sig = SIGNALS_BY_NAME[self._config.deathsig.upper()]
            libc.prctl(libc.PR_SET_PDEATHSIG, sig, 0, 0, 0, 0)


##


RLIMITS_BY_NAME = {
    a[len('RLIMIT_'):]: v  # noqa
    for a in dir(resource)
    if a.startswith('RLIMIT_')
    and a == a.upper()
    and isinstance((v := getattr(resource, a)), int)
}


class RlimitBootstrap(ContextBootstrap['RlimitBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        limits: ta.Optional[ta.Mapping[str, ta.Tuple[ta.Optional[int], ta.Optional[int]]]] = None

    @contextlib.contextmanager
    def enter(self) -> ta.Iterator[None]:
        if not self._config.limits:
            yield
            return

        def or_infin(l: ta.Optional[int]) -> int:
            return l if l is not None else resource.RLIM_INFINITY

        prev = {}
        for k, (s, h) in self._config.limits.items():
            i = RLIMITS_BY_NAME[k.upper()]
            prev[i] = resource.getrlimit(i)
            resource.setrlimit(i, (or_infin(s), or_infin(h)))

        try:
            yield

        finally:
            for i, (s, h) in prev.items():
                resource.setrlimit(i, (s, h))


##


class ImportBootstrap(SimpleBootstrap['ImportBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        modules: ta.Optional[ta.Sequence[str]] = None

    def run(self) -> None:
        for m in self._config.modules or ():
            importlib.import_module(m)


##


class EnvBootstrap(ContextBootstrap['EnvBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        vars: ta.Optional[ta.Mapping[str, ta.Optional[str]]] = None
        files: ta.Optional[ta.Sequence[str]] = None

    @contextlib.contextmanager
    def enter(self) -> ta.Iterator[None]:
        if not (self._config.vars or self._config.files):
            yield
            return

        new = dict(self._config.vars or {})
        for f in self._config.files or ():
            new.update(dotenv.dotenv_values(f, env=os.environ))

        prev: ta.Dict[str, ta.Optional[str]] = {k: os.environ.get(k) for k in new}

        def do(k: str, v: ta.Optional[str]) -> None:
            if v is not None:
                os.environ[k] = v
            else:
                del os.environ[k]

        for k, v in new.items():
            do(k, v)

        try:
            yield

        finally:
            for k, v in prev.items():
                do(k, v)


##


class PidfileBootstrap(ContextBootstrap['PidfileBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        path: ta.Optional[str] = None

    @contextlib.contextmanager
    def enter(self) -> ta.Iterator[None]:
        if self._config.path is None:
            yield
            return

        with pidfile.Pidfile(self._config.path) as pf:
            pf.write()
            yield


##


class FdsBootstrap(SimpleBootstrap['FdsBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        redirects: ta.Optional[ta.Mapping[int, ta.Union[int, str, None]]] = None

    def run(self) -> None:
        for dst, src in (self._config.redirects or {}).items():
            if src is None:
                src = '/dev/null'
            if isinstance(src, int):
                os.dup2(src, dst)
            elif isinstance(src, str):
                sfd = os.open(src, os.O_RDWR)
                os.dup2(sfd, dst)
                os.close(sfd)
            else:
                raise TypeError(src)


##


class PrintPidBootstrap(SimpleBootstrap['PrintPidBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        enable: bool = False
        pause: bool = False

    def run(self) -> None:
        if not (self._config.enable or self._config.pause):
            return
        print(str(os.getpid()), file=sys.stderr)
        if self._config.pause:
            input()



from .abstract import (  # noqa
    AbstractLifecycle,
)

from .base import (  # noqa
    CallbackLifecycle,
    Lifecycle,
    LifecycleCallback,
)

from .contextmanagers import (  # noqa
    ContextManagerLifecycle,
    LifecycleContextManager,
)

from .controller import (  # noqa
    LifecycleController,
    LifecycleListener,
)

from .manager import (  # noqa
    LifecycleManager,
)

from .states import (  # noqa
    LifecycleState,
    LifecycleStateError,
    LifecycleStates,
)

from .transitions import (  # noqa
    LifecycleTransition,
    LifecycleTransitions,
)



"""
This module solely exists for import convenience - tools will tend to resolve the imported 'property' as if it were from
builtins and thus not distinguish it from a normal property.

    from omlish import cached

    class C:
        @cached.property
        def p(self) -> str: ...

"""
from .lang.cached import _CachedProperty  # noqa
from .lang.cached import cached_function

function = cached_function

property = property  # noqa

globals()['property'] = _CachedProperty  # noqa



from ... import dataclasses as dc
from ... import lang
from .exprs import CanExpr
from .exprs import Expr
from .exprs import ExprBuilder
from .ops import OpKind


##


class UnaryOp(dc.Frozen, lang.Final, eq=False):
    name: str
    kind: OpKind


class UnaryOps(lang.Namespace):
    NOT = UnaryOp('not', OpKind.CMP)
    IS_NULL = UnaryOp('is_null', OpKind.CMP)
    IS_NOT_NULL = UnaryOp('is_not_null', OpKind.CMP)

    POS = UnaryOp('pos', OpKind.ARITH)
    NEG = UnaryOp('neg', OpKind.ARITH)


class Unary(Expr, lang.Final):
    op: UnaryOp
    v: Expr


class UnaryBuilder(ExprBuilder):
    def unary(self, op: UnaryOp, v: CanExpr) -> Unary:
        return Unary(op, self.expr(v))

    #

    def not_(self, v: CanExpr) -> Unary:
        return self.unary(UnaryOps.NOT, v)

    def is_null(self, v: CanExpr) -> Unary:
        return self.unary(UnaryOps.IS_NULL, v)

    def is_not_null(self, v: CanExpr) -> Unary:
        return self.unary(UnaryOps.IS_NOT_NULL, v)

    #

    def pos(self, v: CanExpr) -> Unary:
        return self.unary(UnaryOps.POS, v)

    def neg(self, v: CanExpr) -> Unary:
        return self.unary(UnaryOps.NEG, v)



"""
Tools for working with the different async loop implementations - currently anyio, asyncio, trio.

TODO:
 - mark whole class / module?
"""
import abc
import dataclasses as dc
import enum
import functools
import typing as ta

from .. import lang
from .trio_asyncio import check_trio_asyncio
from .trio_asyncio import with_trio_asyncio_loop


if ta.TYPE_CHECKING:
    import sniffio
    import trio_asyncio
else:
    sniffio = lang.proxy_import('sniffio')
    trio_asyncio = lang.proxy_import('trio_asyncio')


T = ta.TypeVar('T')


##


_FLAVOR_ATTR = '__async_flavor__'


class _MISSING(lang.Marker):
    pass


class Flavor(enum.Enum):
    ASYNCIO = enum.auto()
    TRIO = enum.auto()
    ANYIO = enum.auto()


def mark_flavor(f: Flavor):
    if not isinstance(f, Flavor):
        raise TypeError(f)

    def inner(fn):
        setattr(fn, _FLAVOR_ATTR, f)
        return fn

    return inner


mark_asyncio = mark_flavor(Flavor.ASYNCIO)
mark_anyio = mark_flavor(Flavor.ANYIO)
mark_trio = mark_flavor(Flavor.TRIO)

PACKAGE_FLAVORS: ta.MutableMapping[str, Flavor] = {
    'anyio': Flavor.ANYIO,
    'asyncio': Flavor.ASYNCIO,
    'trio': Flavor.TRIO,

    'sqlalchemy.ext.asyncio': Flavor.ASYNCIO,
}

_MODULE_FLAVOR_CACHE: dict[str, Flavor | None] = {}


def _get_module_flavor(p: str) -> Flavor | None:
    try:
        return _MODULE_FLAVOR_CACHE[p]
    except KeyError:
        pass

    pf: Flavor | None = None
    for cp, cf in PACKAGE_FLAVORS.items():
        if p.startswith(cp) and (len(cp) == len(p) or p[len(cp)] == '.'):
            pf = cf
            break

    _MODULE_FLAVOR_CACHE[p] = pf
    return pf


def get_flavor(obj: ta.Any, default: Flavor | type[_MISSING] | None = _MISSING) -> Flavor:
    u = lang.unwrap_func(obj)

    try:
        return getattr(u, _FLAVOR_ATTR)
    except AttributeError:
        pass

    if (mn := getattr(u, '__module__', None)) is not None:
        if (pf := _get_module_flavor(mn)):
            return pf

    if default is not _MISSING:
        return default  # type: ignore

    raise TypeError(f'not marked with flavor: {obj}')


##


def with_adapter_loop(*, wait=False):
    def outer(fn):
        @functools.wraps(fn)
        async def inner(*args, **kwargs):
            cur_lib = sniffio.current_async_library()

            if cur_lib == 'asyncio':
                await fn(*args, **kwargs)

            elif cur_lib == 'trio':
                await with_trio_asyncio_loop(wait=wait)(fn)(*args, **kwargs)

            else:
                raise RuntimeError(f'Unknown async library: {cur_lib}')

        return inner

    return outer


##


class Adapter(lang.Abstract):
    _FROM_METHODS_BY_FLAVOR: ta.ClassVar[ta.Mapping[Flavor, str]] = {
        Flavor.ANYIO: 'from_anyio',
        Flavor.ASYNCIO: 'from_asyncio',
        Flavor.TRIO: 'from_trio',
    }

    def adapt(self, fn, fl=None):
        if fl is None:
            fl = get_flavor(fn)
        return getattr(self, self._FROM_METHODS_BY_FLAVOR[fl])(fn)

    #

    def from_anyio(self, fn):
        return fn

    @abc.abstractmethod
    def from_asyncio(self, fn):
        raise NotImplementedError

    @abc.abstractmethod
    def from_trio(self, fn):
        raise NotImplementedError


class AsyncioAdapter(Adapter):
    def from_asyncio(self, fn):
        return fn

    def from_trio(self, fn):
        check_trio_asyncio()
        return trio_asyncio.trio_as_aio(fn)


class TrioAdapter(Adapter):
    def from_asyncio(self, fn):
        check_trio_asyncio()
        return trio_asyncio.aio_as_trio(fn)

    def from_trio(self, fn):
        return fn


_ADAPTERS_BY_BACKEND: ta.Mapping[str, Adapter] = {
    'asyncio': AsyncioAdapter(),
    'trio': TrioAdapter(),
}


def get_adapter() -> Adapter:
    return _ADAPTERS_BY_BACKEND[sniffio.current_async_library()]


def adapt(fn, fl=None):
    return get_adapter().adapt(fn, fl)


def from_anyio(fn):
    return get_adapter().from_anyio(fn)


def from_asyncio(fn):
    return get_adapter().from_asyncio(fn)


def from_trio(fn):
    return get_adapter().from_trio(fn)


##


@dc.dataclass(frozen=True)
class ContextManagerAdapter(ta.Generic[T]):
    obj: ta.AsyncContextManager[T]
    adapt: ta.Callable[[ta.Callable], ta.Callable]

    async def __aenter__(self, *args: ta.Any, **kwargs: ta.Any) -> T:
        return await self.adapt(self.obj.__aenter__)(*args, **kwargs)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self.adapt(self.obj.__aexit__)(exc_type, exc_val, exc_tb)


def adapt_context(obj):
    return ContextManagerAdapter(obj, get_adapter().adapt)


def from_anyio_context(obj):
    return ContextManagerAdapter(obj, get_adapter().from_anyio)


def from_asyncio_context(obj):
    return ContextManagerAdapter(obj, get_adapter().from_asyncio)


def from_trio_context(obj):
    return ContextManagerAdapter(obj, get_adapter().from_trio)



import typing as ta

from ... import check
from ... import collections as col
from ... import dataclasses as dc
from ... import lang
from .dtypes import Dtype


##


class Element(lang.Abstract, lang.Sealed):
    pass


##


@dc.dataclass(frozen=True)
class Column(Element, lang.Final):
    name: str
    type: Dtype
    nullable: bool = dc.field(default=False, kw_only=True)
    default: lang.Maybe[ta.Any] = dc.field(default=lang.empty(), kw_only=True)


@dc.dataclass(frozen=True)
class PrimaryKey(Element, lang.Final):
    columns: ta.Sequence[str]


@dc.dataclass(frozen=True)
class Index(Element, lang.Final):
    columns: ta.Sequence[str] = dc.xfield(coerce=col.seq)
    name: str | None = None


##


@dc.dataclass(frozen=True)
class IdIntegerPrimaryKey(Element, lang.Final):
    pass


#


@dc.dataclass(frozen=True)
class CreatedAt(Element, lang.Final):
    pass


@dc.dataclass(frozen=True)
class UpdatedAt(Element, lang.Final):
    pass


@dc.dataclass(frozen=True)
class UpdatedAtTrigger(Element, lang.Final):
    column: str


@dc.dataclass(frozen=True)
class CreatedAtUpdatedAt(Element, lang.Final):
    pass


##


@dc.dataclass(frozen=True)
class Elements(ta.Sequence[Element], lang.Final):
    lst: ta.Sequence[Element] = dc.xfield(coerce=col.seq_of(check.of_isinstance(Element)))

    def __iter__(self) -> ta.Iterator[Element]:
        return iter(self.lst)

    def __getitem__(self, index: ta.Any) -> Element:  # type: ignore[override]
        return self.lst[index]

    def __len__(self) -> int:
        return len(self.lst)

    @lang.cached_property
    def by_type(self) -> col.DynamicTypeMap[Element]:
        return col.DynamicTypeMap(self.lst)



"""
TODO:
 - max buf size
 - max recursion depth
 - mark start pos of tokens, currently returning end
"""
import dataclasses as dc
import io
import json
import re
import typing as ta

from .... import check
from ....genmachine import GenMachine


##


ValueTokenKind: ta.TypeAlias = ta.Literal[
    'STRING',
    'NUMBER',

    'SPECIAL_NUMBER',
    'BOOLEAN',
    'NULL',
]

VALUE_TOKEN_KINDS = frozenset(check.isinstance(a, str) for a in ta.get_args(ValueTokenKind))

ControlTokenKind: ta.TypeAlias = ta.Literal[
    'LBRACE',
    'RBRACE',
    'LBRACKET',
    'RBRACKET',
    'COMMA',
    'COLON',
]

TokenKind: ta.TypeAlias = ValueTokenKind | ControlTokenKind

#

ScalarValue: ta.TypeAlias = str | float | int | None

SCALAR_VALUE_TYPES: tuple[type, ...] = tuple(
    check.isinstance(e, type) if e is not None else type(None)
    for e in ta.get_args(ScalarValue)
)


##


class Position(ta.NamedTuple):
    ofs: int
    line: int
    col: int


class Token(ta.NamedTuple):
    kind: TokenKind
    value: ScalarValue
    raw: str | None

    pos: Position

    def __iter__(self):
        raise TypeError


NUMBER_PAT = re.compile(r'-?(?:0|[1-9]\d*)(?:\.\d+)?(?:[eE][+-]?\d+)?')

CONTROL_TOKENS: ta.Mapping[str, TokenKind] = {
    '{': 'LBRACE',
    '}': 'RBRACE',
    '[': 'LBRACKET',
    ']': 'RBRACKET',
    ',': 'COMMA',
    ':': 'COLON',
}

CONST_TOKENS: ta.Mapping[str, tuple[TokenKind, str | float | None]] = {
    'NaN': ('SPECIAL_NUMBER', float('nan')),
    'Infinity': ('SPECIAL_NUMBER', float('inf')),
    '-Infinity': ('SPECIAL_NUMBER', float('-inf')),

    'true': ('BOOLEAN', True),
    'false': ('BOOLEAN', False),
    'null': ('NULL', None),
}


##


@dc.dataclass()
class JsonLexError(Exception):
    message: str

    pos: Position


class JsonStreamLexer(GenMachine[str, Token]):
    def __init__(
            self,
            *,
            include_raw: bool = False,
    ) -> None:
        self._include_raw = include_raw

        self._ofs = 0
        self._line = 1
        self._col = 0

        self._buf = io.StringIO()

        super().__init__(self._do_main())

    @property
    def pos(self) -> Position:
        return Position(
            self._ofs,
            self._line,
            self._col,
        )

    def _char_in(self, c: str) -> str:
        if c and len(c) != 1:
            raise ValueError(c)

        self._ofs += 1

        if c == '\n':
            self._line += 1
            self._col = 0
        else:
            self._col += 1

        return c

    def _make_tok(
            self,
            kind: TokenKind,
            value: ScalarValue,
            raw: str,
            pos: Position,
    ) -> ta.Sequence[Token]:
        tok = Token(
            kind,
            value,
            raw if self._include_raw else None,
            pos,
        )
        return (tok,)

    def _flip_buf(self) -> str:
        raw = self._buf.getvalue()
        self._buf.seek(0)
        self._buf.truncate()
        return raw

    def _raise(self, msg: str) -> ta.NoReturn:
        raise JsonLexError(msg, self.pos)

    def _do_main(self):
        while True:
            c = self._char_in((yield None))  # noqa

            if not c:
                return None

            if c.isspace():
                continue

            if c in CONTROL_TOKENS:
                yield self._make_tok(CONTROL_TOKENS[c], c, c, self.pos)
                continue

            if c == '"':
                return self._do_string()

            if c.isdigit() or c == '-':
                return self._do_number(c)

            if c in 'tfnIN':
                return self._do_const(c)

            self._raise(f'Unexpected character: {c}')

    def _do_string(self):
        check.state(self._buf.tell() == 0)
        self._buf.write('"')

        pos = self.pos

        last = None
        while True:
            try:
                c = self._char_in((yield None))  # noqa
            except GeneratorExit:
                self._raise('Unexpected end of input')

            if not c:
                raise NotImplementedError

            self._buf.write(c)
            if c == '"' and last != '\\':
                break
            last = c

        raw = self._flip_buf()
        sv = json.loads(raw)
        yield self._make_tok('STRING', sv, raw, pos)

        return self._do_main()

    def _do_number(self, c: str):
        check.state(self._buf.tell() == 0)
        self._buf.write(c)

        pos = self.pos

        while True:
            try:
                c = self._char_in((yield None))  # noqa
            except GeneratorExit:
                self._raise('Unexpected end of input')

            if not c:
                break

            if not (c.isdigit() or c in '.eE+-'):
                break
            self._buf.write(c)

        raw = self._flip_buf()

        #

        if not NUMBER_PAT.fullmatch(raw):
            # Can only be -Infinity

            if not c:
                self._raise('Unexpected end of input')

            raw += c
            try:
                for _ in range(7):
                    raw += self._char_in((yield None))  # noqa
            except GeneratorExit:
                self._raise('Unexpected end of input')

            if raw != '-Infinity':
                self._raise(f'Invalid number format: {raw}')

            tk, tv = CONST_TOKENS[raw]
            yield self._make_tok(tk, tv, raw, pos)

            return self._do_main()

        #

        if '.' in raw or 'e' in raw or 'E' in raw:
            nv = float(raw)
        else:
            nv = int(raw)
        yield self._make_tok('NUMBER', nv, raw, pos)

        #

        if not c:
            return None

        if c in CONTROL_TOKENS:
            yield self._make_tok(CONTROL_TOKENS[c], c, c, pos)

        elif not c.isspace():
            self._raise(f'Unexpected character after number: {c}')

        return self._do_main()

    def _do_const(self, c: str):
        pos = self.pos
        raw = c
        while True:
            try:
                raw += self._char_in((yield None))  # noqa
            except GeneratorExit:
                self._raise('Unexpected end of input')

            if raw in CONST_TOKENS:
                break

            if len(raw) > 8:  # None of the keywords are longer than 8 characters
                self._raise(f'Invalid literal: {raw}')

        tk, tv = CONST_TOKENS[raw]
        yield self._make_tok(tk, tv, raw, pos)

        return self._do_main()



"""
TODO:
 - join
 - subquery
"""
import typing as ta

from ... import dataclasses as dc
from ... import lang
from .base import Node
from .idents import Ident
from .names import CanName
from .names import Name
from .names import NameBuilder


##


class Relation(Node, lang.Abstract):
    pass


class Table(Relation, lang.Final):
    n: Name
    a: Ident | None = dc.xfield(None, repr_fn=dc.opt_repr)


CanTable: ta.TypeAlias = Table | CanName
CanRelation: ta.TypeAlias = Relation | CanTable


class RelationBuilder(NameBuilder):
    def table(self, n: CanTable) -> Table:
        if isinstance(n, Table):
            return n
        else:
            return Table(self.name(n))

    def relation(self, o: CanRelation) -> Relation:
        if isinstance(o, Relation):
            return o
        else:
            return self.table(o)



"""
Null
Bool
Int
Float
Number
String
Bytes
Array
Object
Any
"""
import typing as ta


Value = ta.Union[  # noqa
    None,

    bool,
    int,
    float,
    # Number,
    str,
    bytes,

    list,  # list[Value],
    dict,  # dict[str, Value],

    # ta.Any,
]



import typing as ta

from ... import dataclasses as dc
from ... import lang
from .base import Node
from .idents import CanIdent
from .idents import Ident
from .idents import IdentBuilder


##


class Name(Node, lang.Final):
    ps: ta.Sequence[Ident] = dc.xfield(coerce=tuple)


CanName: ta.TypeAlias = Name | CanIdent | ta.Sequence[CanIdent]


class NameBuilder(IdentBuilder):
    def name(self, o: CanName) -> Name:
        if isinstance(o, Name):
            return o
        elif isinstance(o, (Ident, str)):
            return Name([self.ident(o)])
        elif isinstance(o, ta.Sequence):
            return Name([self.ident(p) for p in o])
        else:
            raise TypeError(o)

    @ta.final
    def n(self, o: CanName) -> Name:
        return self.name(o)



"""
TODO:
 - __deepcopy__
"""
import dataclasses as dc

from .fields import field_type
from .internals import FIELDS_ATTR
from .internals import FieldType
from .processing import Processor
from .utils import set_new_attribute


MISSING = dc.MISSING


def _copy(obj):
    kw = {}

    for f in getattr(obj, FIELDS_ATTR).values():
        if field_type(f) is FieldType.CLASS:
            continue
        kw[f.name] = getattr(obj, f.name)

    return obj.__class__(**kw)


class CopyProcessor(Processor):
    def _process(self) -> None:
        set_new_attribute(self._cls, '__copy__', _copy)



# type: ignore
# ruff: noqa
# flake8: noqa
#
# Copyright (c) 2012-2017 The ANTLR Project. All rights reserved.
# Use of this file is governed by the BSD 3-clause license that
# can be found in the LICENSE.txt file in the project root.
#/
from .RuleContext import RuleContext
from .atn.ATN import ATN
from .error.Errors import IllegalStateException
from io import StringIO

# dup ParserATNSimulator class var here to avoid circular import; no idea why this can't be in PredictionContext
_trace_atn_sim = False

class PredictionContext(object):

    # Represents {@code $} in local context prediction, which means wildcard.
    # {@code#+x =#}.
    #/
    EMPTY = None

    # Represents {@code $} in an array in full context mode, when {@code $}
    # doesn't mean wildcard: {@code $ + x = [$,x]}. Here,
    # {@code $} = {@link #EMPTY_RETURN_STATE}.
    #/
    EMPTY_RETURN_STATE = 0x7FFFFFFF

    globalNodeCount = 1
    id = globalNodeCount

    # Stores the computed hash code of this {@link PredictionContext}. The hash
    # code is computed in parts to match the following reference algorithm.
    #
    # <pre>
    #  private int referenceHashCode() {
    #      int hash = {@link MurmurHash#initialize MurmurHash.initialize}({@link #INITIAL_HASH});
    #
    #      for (int i = 0; i &lt; {@link #size()}; i++) {
    #          hash = {@link MurmurHash#update MurmurHash.update}(hash, {@link #getParent getParent}(i));
    #      }
    #
    #      for (int i = 0; i &lt; {@link #size()}; i++) {
    #          hash = {@link MurmurHash#update MurmurHash.update}(hash, {@link #getReturnState getReturnState}(i));
    #      }
    #
    #      hash = {@link MurmurHash#finish MurmurHash.finish}(hash, 2# {@link #size()});
    #      return hash;
    #  }
    # </pre>
    #/

    def __init__(self, cachedHashCode:int):
        self.cachedHashCode = cachedHashCode

    def __len__(self):
        return 0

    # This means only the {@link #EMPTY} context is in set.
    def isEmpty(self):
        return self is self.EMPTY

    def hasEmptyPath(self):
        return self.getReturnState(len(self) - 1) == self.EMPTY_RETURN_STATE

    def getReturnState(self, index:int):
        raise IllegalStateException("illegal!")

    def __hash__(self):
        return self.cachedHashCode

def calculateHashCode(parent:PredictionContext, returnState:int):
    return hash("") if parent is None else hash((hash(parent), returnState))

def calculateListsHashCode(parents:[], returnStates:[] ):
    h = 0
    for parent, returnState in zip(parents, returnStates):
        h = hash((h, calculateHashCode(parent, returnState)))
    return h

#  Used to cache {@link PredictionContext} objects. Its used for the shared
#  context cash associated with contexts in DFA states. This cache
#  can be used for both lexers and parsers.

class PredictionContextCache(object):

    def __init__(self):
        self.cache = dict()

    #  Add a context to the cache and return it. If the context already exists,
    #  return that one instead and do not add a new context to the cache.
    #  Protect shared cache from unsafe thread access.
    #
    def add(self, ctx:PredictionContext):
        if ctx==PredictionContext.EMPTY:
            return PredictionContext.EMPTY
        existing = self.cache.get(ctx, None)
        if existing is not None:
            return existing
        self.cache[ctx] = ctx
        return ctx

    def get(self, ctx:PredictionContext):
        return self.cache.get(ctx, None)

    def __len__(self):
        return len(self.cache)


class SingletonPredictionContext(PredictionContext):

    @staticmethod
    def create(parent:PredictionContext , returnState:int ):
        if returnState == PredictionContext.EMPTY_RETURN_STATE and parent is None:
            # someone can pass in the bits of an array ctx that mean $
            return SingletonPredictionContext.EMPTY
        else:
            return SingletonPredictionContext(parent, returnState)

    def __init__(self, parent:PredictionContext, returnState:int):
        hashCode = calculateHashCode(parent, returnState)
        super().__init__(hashCode)
        self.parentCtx = parent
        self.returnState = returnState

    def __len__(self):
        return 1

    def getParent(self, index:int):
        return self.parentCtx

    def getReturnState(self, index:int):
        return self.returnState

    def __eq__(self, other):
        if self is other:
            return True
        elif other is None:
            return False
        elif not isinstance(other, SingletonPredictionContext):
            return False
        else:
            return self.returnState == other.returnState and self.parentCtx == other.parentCtx

    def __hash__(self):
        return self.cachedHashCode

    def __str__(self):
        up = "" if self.parentCtx is None else str(self.parentCtx)
        if len(up)==0:
            if self.returnState == self.EMPTY_RETURN_STATE:
                return "$"
            else:
                return str(self.returnState)
        else:
            return str(self.returnState) + " " + up


class EmptyPredictionContext(SingletonPredictionContext):

    def __init__(self):
        super().__init__(None, PredictionContext.EMPTY_RETURN_STATE)

    def isEmpty(self):
        return True

    def __eq__(self, other):
        return self is other

    def __hash__(self):
        return self.cachedHashCode

    def __str__(self):
        return "$"


PredictionContext.EMPTY = EmptyPredictionContext()

class ArrayPredictionContext(PredictionContext):
    # Parent can be null only if full ctx mode and we make an array
    #  from {@link #EMPTY} and non-empty. We merge {@link #EMPTY} by using null parent and
    #  returnState == {@link #EMPTY_RETURN_STATE}.

    def __init__(self, parents:list, returnStates:list):
        super().__init__(calculateListsHashCode(parents, returnStates))
        self.parents = parents
        self.returnStates = returnStates

    def isEmpty(self):
        # since EMPTY_RETURN_STATE can only appear in the last position, we
        # don't need to verify that size==1
        return self.returnStates[0]==PredictionContext.EMPTY_RETURN_STATE

    def __len__(self):
        return len(self.returnStates)

    def getParent(self, index:int):
        return self.parents[index]

    def getReturnState(self, index:int):
        return self.returnStates[index]

    def __eq__(self, other):
        if self is other:
            return True
        elif not isinstance(other, ArrayPredictionContext):
            return False
        elif hash(self) != hash(other):
            return False # can't be same if hash is different
        else:
            return self.returnStates==other.returnStates and self.parents==other.parents

    def __str__(self):
        if self.isEmpty():
            return "[]"
        with StringIO() as buf:
            buf.write("[")
            for i in range(0,len(self.returnStates)):
                if i>0:
                    buf.write(", ")
                if self.returnStates[i]==PredictionContext.EMPTY_RETURN_STATE:
                    buf.write("$")
                    continue
                buf.write(str(self.returnStates[i]))
                if self.parents[i] is not None:
                    buf.write(' ')
                    buf.write(str(self.parents[i]))
                else:
                    buf.write("null")
            buf.write("]")
            return buf.getvalue()

    def __hash__(self):
        return self.cachedHashCode



#  Convert a {@link RuleContext} tree to a {@link PredictionContext} graph.
#  Return {@link #EMPTY} if {@code outerContext} is empty or null.
#/
def PredictionContextFromRuleContext(atn:ATN, outerContext:RuleContext=None):
    if outerContext is None:
        outerContext = RuleContext.EMPTY

    # if we are in RuleContext of start rule, s, then PredictionContext
    # is EMPTY. Nobody called us. (if we are empty, return empty)
    if outerContext.parentCtx is None or outerContext is RuleContext.EMPTY:
        return PredictionContext.EMPTY

    # If we have a parent, convert it to a PredictionContext graph
    parent = PredictionContextFromRuleContext(atn, outerContext.parentCtx)
    state = atn.states[outerContext.invokingState]
    transition = state.transitions[0]
    return SingletonPredictionContext.create(parent, transition.followState.stateNumber)


def merge(a:PredictionContext, b:PredictionContext, rootIsWildcard:bool, mergeCache:dict):

    # share same graph if both same
    if a==b:
        return a

    if isinstance(a, SingletonPredictionContext) and isinstance(b, SingletonPredictionContext):
        return mergeSingletons(a, b, rootIsWildcard, mergeCache)

    # At least one of a or b is array
    # If one is $ and rootIsWildcard, return $ as# wildcard
    if rootIsWildcard:
        if isinstance( a, EmptyPredictionContext ):
            return a
        if isinstance( b, EmptyPredictionContext ):
            return b

    # convert singleton so both are arrays to normalize
    if isinstance( a, SingletonPredictionContext ):
        a = ArrayPredictionContext([a.parentCtx], [a.returnState])
    if isinstance( b, SingletonPredictionContext):
        b = ArrayPredictionContext([b.parentCtx], [b.returnState])
    return mergeArrays(a, b, rootIsWildcard, mergeCache)


#
# Merge two {@link SingletonPredictionContext} instances.
#
# <p>Stack tops equal, parents merge is same; return left graph.<br>
# <embed src="images/SingletonMerge_SameRootSamePar.svg" type="image/svg+xml"/></p>
#
# <p>Same stack top, parents differ; merge parents giving array node, then
# remainders of those graphs. A new root node is created to point to the
# merged parents.<br>
# <embed src="images/SingletonMerge_SameRootDiffPar.svg" type="image/svg+xml"/></p>
#
# <p>Different stack tops pointing to same parent. Make array node for the
# root where both element in the root point to the same (original)
# parent.<br>
# <embed src="images/SingletonMerge_DiffRootSamePar.svg" type="image/svg+xml"/></p>
#
# <p>Different stack tops pointing to different parents. Make array node for
# the root where each element points to the corresponding original
# parent.<br>
# <embed src="images/SingletonMerge_DiffRootDiffPar.svg" type="image/svg+xml"/></p>
#
# @param a the first {@link SingletonPredictionContext}
# @param b the second {@link SingletonPredictionContext}
# @param rootIsWildcard {@code true} if this is a local-context merge,
# otherwise false to indicate a full-context merge
# @param mergeCache
#/
def mergeSingletons(a:SingletonPredictionContext, b:SingletonPredictionContext, rootIsWildcard:bool, mergeCache:dict):
    if mergeCache is not None:
        previous = mergeCache.get((a,b), None)
        if previous is not None:
            return previous
        previous = mergeCache.get((b,a), None)
        if previous is not None:
            return previous

    merged = mergeRoot(a, b, rootIsWildcard)
    if merged is not None:
        if mergeCache is not None:
            mergeCache[(a, b)] = merged
        return merged

    if a.returnState==b.returnState:
        parent = merge(a.parentCtx, b.parentCtx, rootIsWildcard, mergeCache)
        # if parent is same as existing a or b parent or reduced to a parent, return it
        if parent == a.parentCtx:
            return a # ax + bx = ax, if a=b
        if parent == b.parentCtx:
            return b # ax + bx = bx, if a=b
        # else: ax + ay = a'[x,y]
        # merge parents x and y, giving array node with x,y then remainders
        # of those graphs.  dup a, a' points at merged array
        # new joined parent so create new singleton pointing to it, a'
        merged = SingletonPredictionContext.create(parent, a.returnState)
        if mergeCache is not None:
            mergeCache[(a, b)] = merged
        return merged
    else: # a != b payloads differ
        # see if we can collapse parents due to $+x parents if local ctx
        singleParent = None
        if a is b or (a.parentCtx is not None and a.parentCtx==b.parentCtx): # ax + bx = [a,b]x
            singleParent = a.parentCtx
        if singleParent is not None:	# parents are same
            # sort payloads and use same parent
            payloads = [ a.returnState, b.returnState ]
            if a.returnState > b.returnState:
                payloads = [ b.returnState, a.returnState ]
            parents = [singleParent, singleParent]
            merged = ArrayPredictionContext(parents, payloads)
            if mergeCache is not None:
                mergeCache[(a, b)] = merged
            return merged
        # parents differ and can't merge them. Just pack together
        # into array; can't merge.
        # ax + by = [ax,by]
        payloads = [ a.returnState, b.returnState ]
        parents = [ a.parentCtx, b.parentCtx ]
        if a.returnState > b.returnState: # sort by payload
            payloads = [ b.returnState, a.returnState ]
            parents = [ b.parentCtx, a.parentCtx ]
        merged = ArrayPredictionContext(parents, payloads)
        if mergeCache is not None:
            mergeCache[(a, b)] = merged
        return merged


#
# Handle case where at least one of {@code a} or {@code b} is
# {@link #EMPTY}. In the following diagrams, the symbol {@code $} is used
# to represent {@link #EMPTY}.
#
# <h2>Local-Context Merges</h2>
#
# <p>These local-context merge operations are used when {@code rootIsWildcard}
# is true.</p>
#
# <p>{@link #EMPTY} is superset of any graph; return {@link #EMPTY}.<br>
# <embed src="images/LocalMerge_EmptyRoot.svg" type="image/svg+xml"/></p>
#
# <p>{@link #EMPTY} and anything is {@code #EMPTY}, so merged parent is
# {@code #EMPTY}; return left graph.<br>
# <embed src="images/LocalMerge_EmptyParent.svg" type="image/svg+xml"/></p>
#
# <p>Special case of last merge if local context.<br>
# <embed src="images/LocalMerge_DiffRoots.svg" type="image/svg+xml"/></p>
#
# <h2>Full-Context Merges</h2>
#
# <p>These full-context merge operations are used when {@code rootIsWildcard}
# is false.</p>
#
# <p><embed src="images/FullMerge_EmptyRoots.svg" type="image/svg+xml"/></p>
#
# <p>Must keep all contexts; {@link #EMPTY} in array is a special value (and
# null parent).<br>
# <embed src="images/FullMerge_EmptyRoot.svg" type="image/svg+xml"/></p>
#
# <p><embed src="images/FullMerge_SameRoot.svg" type="image/svg+xml"/></p>
#
# @param a the first {@link SingletonPredictionContext}
# @param b the second {@link SingletonPredictionContext}
# @param rootIsWildcard {@code true} if this is a local-context merge,
# otherwise false to indicate a full-context merge
#/
def mergeRoot(a:SingletonPredictionContext, b:SingletonPredictionContext, rootIsWildcard:bool):
    if rootIsWildcard:
        if a == PredictionContext.EMPTY:
            return PredictionContext.EMPTY  ## + b =#
        if b == PredictionContext.EMPTY:
            return PredictionContext.EMPTY  # a +# =#
    else:
        if a == PredictionContext.EMPTY and b == PredictionContext.EMPTY:
            return PredictionContext.EMPTY # $ + $ = $
        elif a == PredictionContext.EMPTY: # $ + x = [$,x]
            payloads = [ b.returnState, PredictionContext.EMPTY_RETURN_STATE ]
            parents = [ b.parentCtx, None ]
            return ArrayPredictionContext(parents, payloads)
        elif b == PredictionContext.EMPTY: # x + $ = [$,x] ($ is always first if present)
            payloads = [ a.returnState, PredictionContext.EMPTY_RETURN_STATE ]
            parents = [ a.parentCtx, None ]
            return ArrayPredictionContext(parents, payloads)
    return None


#
# Merge two {@link ArrayPredictionContext} instances.
#
# <p>Different tops, different parents.<br>
# <embed src="images/ArrayMerge_DiffTopDiffPar.svg" type="image/svg+xml"/></p>
#
# <p>Shared top, same parents.<br>
# <embed src="images/ArrayMerge_ShareTopSamePar.svg" type="image/svg+xml"/></p>
#
# <p>Shared top, different parents.<br>
# <embed src="images/ArrayMerge_ShareTopDiffPar.svg" type="image/svg+xml"/></p>
#
# <p>Shared top, all shared parents.<br>
# <embed src="images/ArrayMerge_ShareTopSharePar.svg" type="image/svg+xml"/></p>
#
# <p>Equal tops, merge parents and reduce top to
# {@link SingletonPredictionContext}.<br>
# <embed src="images/ArrayMerge_EqualTop.svg" type="image/svg+xml"/></p>
#/
def mergeArrays(a:ArrayPredictionContext, b:ArrayPredictionContext, rootIsWildcard:bool, mergeCache:dict):
    if mergeCache is not None:
        previous = mergeCache.get((a,b), None)
        if previous is not None:
            if _trace_atn_sim: print("mergeArrays a="+str(a)+",b="+str(b)+" -> previous")
            return previous
        previous = mergeCache.get((b,a), None)
        if previous is not None:
            if _trace_atn_sim: print("mergeArrays a="+str(a)+",b="+str(b)+" -> previous")
            return previous

    # merge sorted payloads a + b => M
    i = 0 # walks a
    j = 0 # walks b
    k = 0 # walks target M array

    mergedReturnStates = [None] * (len(a.returnStates) + len( b.returnStates))
    mergedParents = [None] * len(mergedReturnStates)
    # walk and merge to yield mergedParents, mergedReturnStates
    while i<len(a.returnStates) and j<len(b.returnStates):
        a_parent = a.parents[i]
        b_parent = b.parents[j]
        if a.returnStates[i]==b.returnStates[j]:
            # same payload (stack tops are equal), must yield merged singleton
            payload = a.returnStates[i]
            # $+$ = $
            bothDollars = payload == PredictionContext.EMPTY_RETURN_STATE and \
                            a_parent is None and b_parent is None
            ax_ax = (a_parent is not None and b_parent is not None) and a_parent==b_parent # ax+ax -> ax
            if bothDollars or ax_ax:
                mergedParents[k] = a_parent # choose left
                mergedReturnStates[k] = payload
            else: # ax+ay -> a'[x,y]
                mergedParent = merge(a_parent, b_parent, rootIsWildcard, mergeCache)
                mergedParents[k] = mergedParent
                mergedReturnStates[k] = payload
            i += 1 # hop over left one as usual
            j += 1 # but also skip one in right side since we merge
        elif a.returnStates[i]<b.returnStates[j]: # copy a[i] to M
            mergedParents[k] = a_parent
            mergedReturnStates[k] = a.returnStates[i]
            i += 1
        else: # b > a, copy b[j] to M
            mergedParents[k] = b_parent
            mergedReturnStates[k] = b.returnStates[j]
            j += 1
        k += 1

    # copy over any payloads remaining in either array
    if i < len(a.returnStates):
        for p in range(i, len(a.returnStates)):
            mergedParents[k] = a.parents[p]
            mergedReturnStates[k] = a.returnStates[p]
            k += 1
    else:
        for p in range(j, len(b.returnStates)):
            mergedParents[k] = b.parents[p]
            mergedReturnStates[k] = b.returnStates[p]
            k += 1

    # trim merged if we combined a few that had same stack tops
    if k < len(mergedParents): # write index < last position; trim
        if k == 1: # for just one merged element, return singleton top
            merged = SingletonPredictionContext.create(mergedParents[0], mergedReturnStates[0])
            if mergeCache is not None:
                mergeCache[(a,b)] = merged
            return merged
        mergedParents = mergedParents[0:k]
        mergedReturnStates = mergedReturnStates[0:k]

    merged = ArrayPredictionContext(mergedParents, mergedReturnStates)

    # if we created same array as a or b, return that instead
    # TODO: track whether this is possible above during merge sort for speed
    if merged==a:
        if mergeCache is not None:
            mergeCache[(a,b)] = a
        if _trace_atn_sim: print("mergeArrays a="+str(a)+",b="+str(b)+" -> a")
        return a
    if merged==b:
        if mergeCache is not None:
            mergeCache[(a,b)] = b
        if _trace_atn_sim: print("mergeArrays a="+str(a)+",b="+str(b)+" -> b")
        return b
    combineCommonParents(mergedParents)

    if mergeCache is not None:
        mergeCache[(a,b)] = merged

    if _trace_atn_sim: print("mergeArrays a="+str(a)+",b="+str(b)+" -> "+str(M))

    return merged


#
# Make pass over all <em>M</em> {@code parents}; merge any {@code equals()}
# ones.
#/
def combineCommonParents(parents:list):
    uniqueParents = dict()

    for p in range(0, len(parents)):
        parent = parents[p]
        if uniqueParents.get(parent, None) is None:
            uniqueParents[parent] = parent

    for p in range(0, len(parents)):
        parents[p] = uniqueParents[parents[p]]

def getCachedPredictionContext(context:PredictionContext, contextCache:PredictionContextCache, visited:dict):
    if context.isEmpty():
        return context
    existing = visited.get(context)
    if existing is not None:
        return existing
    existing = contextCache.get(context)
    if existing is not None:
        visited[context] = existing
        return existing
    changed = False
    parents = [None] * len(context)
    for i in range(0, len(parents)):
        parent = getCachedPredictionContext(context.getParent(i), contextCache, visited)
        if changed or parent is not context.getParent(i):
            if not changed:
                parents = [context.getParent(j) for j in range(len(context))]
                changed = True
            parents[i] = parent
    if not changed:
        contextCache.add(context)
        visited[context] = context
        return context

    updated = None
    if len(parents) == 0:
        updated = PredictionContext.EMPTY
    elif len(parents) == 1:
        updated = SingletonPredictionContext.create(parents[0], context.getReturnState(0))
    else:
        updated = ArrayPredictionContext(parents, context.returnStates)

    contextCache.add(updated)
    visited[updated] = updated
    visited[context] = updated

    return updated


#	# extra structures, but cut/paste/morphed works, so leave it.
#	# seems to do a breadth-first walk
#	public static List<PredictionContext> getAllNodes(PredictionContext context) {
#		Map<PredictionContext, PredictionContext> visited =
#			new IdentityHashMap<PredictionContext, PredictionContext>();
#		Deque<PredictionContext> workList = new ArrayDeque<PredictionContext>();
#		workList.add(context);
#		visited.put(context, context);
#		List<PredictionContext> nodes = new ArrayList<PredictionContext>();
#		while (!workList.isEmpty()) {
#			PredictionContext current = workList.pop();
#			nodes.add(current);
#			for (int i = 0; i < current.size(); i++) {
#				PredictionContext parent = current.getParent(i);
#				if ( parent!=null && visited.put(parent, parent) == null) {
#					workList.push(parent);
#				}
#			}
#		}
#		return nodes;
#	}

# ter's recursive version of Sam's getAllNodes()
def getAllContextNodes(context:PredictionContext, nodes:list=None, visited:dict=None):
    if nodes is None:
        nodes = list()
        return getAllContextNodes(context, nodes, visited)
    elif visited is None:
        visited = dict()
        return getAllContextNodes(context, nodes, visited)
    else:
        if context is None or visited.get(context, None) is not None:
            return nodes
        visited.put(context, context)
        nodes.add(context)
        for i in range(0, len(context)):
            getAllContextNodes(context.getParent(i), nodes, visited)
        return nodes




"""
TODO:
 - objects
  - csv
  - csvloader
 - wrapped (wait for usecase)
 - streams / incremental
  - fileobj -> fileobj?
 - swap zstandard for zstd

Compression choice:
 - lzma if-available minimal-space
 - lz4 if-available write-once
 - zstd if-available
 - bz2 write-once (but no parallel decompress)
 - gz
"""
import abc
import codecs
import dataclasses as dc
import typing as ta

from . import lang


if ta.TYPE_CHECKING:
    import bz2 as _bz2
    import gzip as _gzip
    import lzma as _lzma
    import pickle as _pickle
    import struct as _struct
    import tomllib as _tomllib

    import cbor2 as _cbor2
    import cloudpickle as _cloudpickle
    import json5 as _json5
    import lz4.frame as _lz4_frame
    import snappy as _snappy
    import yaml as _yaml
    import zstandard as _zstandard

    from .formats import json as _json

else:
    _bz2 = lang.proxy_import('bz2')
    _gzip = lang.proxy_import('gzip')
    _lzma = lang.proxy_import('lzma')
    _pickle = lang.proxy_import('pickle')
    _struct = lang.proxy_import('struct')
    _tomllib = lang.proxy_import('tomllib')

    _cbor2 = lang.proxy_import('cbor2')
    _cloudpickle = lang.proxy_import('cloudpickle')
    _json5 = lang.proxy_import('json5')
    _lz4_frame = lang.proxy_import('lz4.frame')
    _snappy = lang.proxy_import('snappy')
    _yaml = lang.proxy_import('yaml')
    _zstandard = lang.proxy_import('zstandard')

    _json = lang.proxy_import('.formats.json', __package__)


##


F = ta.TypeVar('F')
T = ta.TypeVar('T')
U = ta.TypeVar('U')


class FnPair(ta.Generic[F, T], abc.ABC):
    @abc.abstractmethod
    def forward(self, f: F) -> T:
        raise NotImplementedError

    @abc.abstractmethod
    def backward(self, t: T) -> F:
        raise NotImplementedError

    ##

    def __call__(self, f: F) -> T:
        return self.forward(f)

    def invert(self) -> 'FnPair[T, F]':
        if isinstance(self, Inverted):
            return self.fp
        return Inverted(self)

    def compose(self, nxt: 'FnPair[T, U]') -> 'FnPair[F, U]':
        return Composite((self, nxt))


##


@lang.unabstract_class(['forward', 'backward'])
@dc.dataclass(frozen=True)
class Simple(FnPair[F, T]):
    forward: ta.Callable[[F], T]  # type: ignore
    backward: ta.Callable[[T], F]  # type: ignore


of = Simple

NOP: FnPair[ta.Any, ta.Any] = of(lang.identity, lang.identity)


##


@dc.dataclass(frozen=True)
class Inverted(FnPair[F, T]):
    fp: FnPair[T, F]

    def forward(self, f: F) -> T:
        return self.fp.backward(f)

    def backward(self, t: T) -> F:
        return self.fp.forward(t)


##


@dc.dataclass(frozen=True)
class Composite(FnPair[F, T]):
    children: ta.Sequence[FnPair]

    def forward(self, f: F) -> T:
        for c in self.children:
            f = c.forward(f)
        return ta.cast(T, f)

    def backward(self, t: T) -> F:
        for c in reversed(self.children):
            t = c.backward(t)
        return ta.cast(F, t)


I0 = ta.TypeVar('I0')
I1 = ta.TypeVar('I1')
I2 = ta.TypeVar('I2')
I3 = ta.TypeVar('I3')
I4 = ta.TypeVar('I4')


@ta.overload
def compose(
        fp0: FnPair[F, I0],
        f01: FnPair[I0, T],
) -> FnPair[F, T]:
    ...


@ta.overload
def compose(
        fp0: FnPair[F, I0],
        f01: FnPair[I0, I1],
        fp2: FnPair[I1, T],
) -> FnPair[F, T]:
    ...


@ta.overload
def compose(
        fp0: FnPair[F, I0],
        f01: FnPair[I0, I1],
        fp2: FnPair[I1, I2],
        fp3: FnPair[I2, T],
) -> FnPair[F, T]:
    ...


@ta.overload
def compose(
        fp0: FnPair[F, I0],
        f01: FnPair[I0, I1],
        fp2: FnPair[I1, I2],
        fp3: FnPair[I2, I3],
        fp4: FnPair[I3, T],
) -> FnPair[F, T]:
    ...


@ta.overload
def compose(
        fp0: FnPair[F, I0],
        f01: FnPair[I0, I1],
        fp2: FnPair[I1, I2],
        fp3: FnPair[I2, I3],
        fp4: FnPair[I3, I4],
        fp5: FnPair[I4, T],
) -> FnPair[F, T]:
    ...


@ta.overload
def compose(*ps: FnPair) -> FnPair:
    ...


def compose(*ps):
    if not ps:
        return NOP
    if len(ps) == 1:
        return ps[0]
    return Composite(ps)


##


@dc.dataclass(frozen=True)
class Text(FnPair[str, bytes]):
    ci: codecs.CodecInfo
    encode_errors: str = dc.field(default='strict', kw_only=True)
    decode_errors: str = dc.field(default='strict', kw_only=True)

    def forward(self, f: str) -> bytes:
        # Python ignores the returned length:
        #   https://github.com/python/cpython/blob/7431c3799efbd06ed03ee70b64420f45e83b3667/Python/codecs.c#L424
        t, _ = self.ci.encode(f, self.encode_errors)
        return t

    def backward(self, t: bytes) -> str:
        f, _ = self.ci.decode(t, self.decode_errors)
        return f


def text(name: str, *, encode_errors: str = 'strict', decode_errors: str = 'strict') -> Text:
    ci = codecs.lookup(name)
    if not ci._is_text_encoding:  # noqa
        raise TypeError(f'must be text codec: {name}')
    return Text(ci, encode_errors=encode_errors, decode_errors=decode_errors)


UTF8 = text('utf-8')


#


@dc.dataclass(frozen=True)
class Optional(FnPair[F | None, T | None]):
    fp: FnPair[F, T]

    def forward(self, f: F | None) -> T | None:
        return None if f is None else self.fp.forward(f)

    def backward(self, t: T | None) -> F | None:
        return None if t is None else self.fp.backward(t)


class Lines(FnPair[ta.Sequence[str], str]):
    def forward(self, f: ta.Sequence[str]) -> str:
        return '\n'.join(f)

    def backward(self, t: str) -> ta.Sequence[str]:
        return t.splitlines()


##


_EXTENSION_REGISTRY: dict[str, type[FnPair]] = {}


def _register_extension(*ss):
    def inner(cls):
        for s in ss:
            if s in _EXTENSION_REGISTRY:
                raise KeyError(s)
            _EXTENSION_REGISTRY[s] = cls
        return cls
    return inner


def get_for_extension(ext: str) -> FnPair:
    return compose(*[_EXTENSION_REGISTRY[p]() for p in ext.split('.')])


##


class Compression(FnPair[bytes, bytes], abc.ABC):
    pass


@_register_extension('bz2')
@dc.dataclass(frozen=True)
class Bz2(Compression):
    compresslevel: int = 9

    def forward(self, f: bytes) -> bytes:
        return _bz2.compress(f, compresslevel=self.compresslevel)

    def backward(self, t: bytes) -> bytes:
        return _bz2.decompress(t)


@_register_extension('gz')
@dc.dataclass(frozen=True)
class Gzip(Compression):
    compresslevel: int = 9

    def forward(self, f: bytes) -> bytes:
        return _gzip.compress(f, compresslevel=self.compresslevel)

    def backward(self, t: bytes) -> bytes:
        return _gzip.decompress(t)


@_register_extension('lzma')
class Lzma(Compression):
    def forward(self, f: bytes) -> bytes:
        return _lzma.compress(f)

    def backward(self, t: bytes) -> bytes:
        return _lzma.decompress(t)


#


@_register_extension('lz4')
@dc.dataclass(frozen=True)
class Lz4(Compression):
    compression_level: int = 0

    def forward(self, f: bytes) -> bytes:
        return _lz4_frame.compress(f, compression_level=self.compression_level)

    def backward(self, t: bytes) -> bytes:
        return _lz4_frame.decompress(t)


@_register_extension('snappy')
class Snappy(Compression):
    def forward(self, f: bytes) -> bytes:
        return _snappy.compress(f)

    def backward(self, t: bytes) -> bytes:
        return _snappy.decompress(t)


@_register_extension('zstd')
class Zstd(Compression):
    def forward(self, f: bytes) -> bytes:
        return _zstandard.compress(f)

    def backward(self, t: bytes) -> bytes:
        return _zstandard.decompress(t)


##


@dc.dataclass(frozen=True)
class Struct(FnPair[tuple, bytes]):
    fmt: str

    def forward(self, f: tuple) -> bytes:
        return _struct.pack(self.fmt, *f)

    def backward(self, t: bytes) -> tuple:
        return _struct.unpack(self.fmt, t)


##


Object: ta.TypeAlias = FnPair[ta.Any, T]
ObjectStr: ta.TypeAlias = Object[str]
ObjectBytes: ta.TypeAlias = Object[bytes]


class Object_(FnPair[ta.Any, T], lang.Abstract):  # noqa
    pass


class ObjectStr_(Object_[str], lang.Abstract):  # noqa
    pass


class ObjectBytes_(Object_[bytes], lang.Abstract):  # noqa
    pass


#


@_register_extension('pkl')
@dc.dataclass(frozen=True)
class Pickle(ObjectBytes_):
    protocol: int | None = None

    def forward(self, f: ta.Any) -> bytes:
        return _pickle.dumps(f, protocol=self.protocol)

    def backward(self, t: bytes) -> ta.Any:
        return _pickle.loads(t)


class _Json(ObjectStr_, lang.Abstract):  # noqa
    def backward(self, t: str) -> ta.Any:
        return _json.loads(t)


@_register_extension('json')
class Json(_Json):
    def forward(self, f: ta.Any) -> str:
        return _json.dumps(f)


class JsonPretty(_Json):
    def forward(self, f: ta.Any) -> str:
        return _json.dumps_pretty(f)


class JsonCompact(_Json):
    def forward(self, f: ta.Any) -> str:
        return _json.dumps_compact(f)


JSON = Json()
PRETTY_JSON = JsonPretty()
COMPACT_JSON = JsonCompact()


@_register_extension('jsonl')
class JsonLines(FnPair[ta.Sequence[ta.Any], str]):
    def forward(self, f: ta.Sequence[ta.Any]) -> str:
        return '\n'.join(_json.dumps(e) for e in f)

    def backward(self, t: str) -> ta.Sequence[ta.Any]:
        return [_json.loads(l) for l in t.splitlines()]


@_register_extension('toml')
class Toml(ObjectStr_):
    def forward(self, f: ta.Any) -> str:
        raise NotImplementedError

    def backward(self, t: str) -> ta.Any:
        return _tomllib.loads(t)


#


@_register_extension('cbor')
class Cbor(ObjectBytes_):
    def forward(self, f: ta.Any) -> bytes:
        return _cbor2.dumps(f)

    def backward(self, t: bytes) -> ta.Any:
        return _cbor2.loads(t)


@_register_extension('clpkl')
@dc.dataclass(frozen=True)
class Cloudpickle(ObjectBytes_):
    protocol: int | None = None

    def forward(self, f: ta.Any) -> bytes:
        return _cloudpickle.dumps(f, protocol=self.protocol)

    def backward(self, t: bytes) -> ta.Any:
        return _cloudpickle.loads(t)


@_register_extension('json5')
class Json5(ObjectStr_):
    def forward(self, f: ta.Any) -> str:
        return _json5.dumps(f)

    def backward(self, t: str) -> ta.Any:
        return _json5.loads(t)


@_register_extension('yml', 'yaml')
class Yaml(ObjectStr_):
    def forward(self, f: ta.Any) -> str:
        return _yaml.dump(f)

    def backward(self, t: str) -> ta.Any:
        return _yaml.safe_load(t)


class YamlUnsafe(ObjectStr_):
    def forward(self, f: ta.Any) -> str:
        return _yaml.dump(f)

    def backward(self, t: str) -> ta.Any:
        return _yaml.load(t, _yaml.FullLoader)



# """
# TODO:
#  - -I/-Ogz, lz4, etc
#   - fnpairs? or not yet just do it
# """
# import abc
# import contextlib
# import dataclasses as dc
# import gzip
# import os
# import typing as ta
#
# from .... import lang
#
#
# if ta.TYPE_CHECKING:
#     import bz2 as _bz2
#     import gzip as _gzip
#     import lzma as _lzma
# else:
#     _bz2 = lang.proxy_import('bz2')
#     _gzip = lang.proxy_import('gzip')
#     _lzma = lang.proxy_import('lzma')
#
#
# ##
#
#
# class BaseIo(lang.Abstract):
#     def close(self) -> None:
#         raise NotImplementedError
#
#     def fileno(self) -> int | None:
#         return None
#
#
# class Input(BaseIo):
#     @abc.abstractmethod
#     def read(self, sz: int | None = None) -> bytes:
#         raise NotImplementedError
#
#
# class Output(BaseIo):
#     @abc.abstractmethod
#     def write(self, data: bytes) -> int:
#         raise NotImplementedError
#
#
# #
#
#
# DEFAULT_READ_SZ = 0x4000
#
#
# @dc.dataclass(frozen=True)
# class FdIo(Input, Output):
#     fd: int
#
#     default_read_sz: int = DEFAULT_READ_SZ
#
#     def read(self, sz: int | None = None) -> bytes:
#         return os.read(self.fd, sz or self.default_read_sz)
#
#     def write(self, data: bytes) -> int:
#         return os.write(self.fd, data)
#
#
# ##
#
#
# @contextlib.contextmanager
# def gzip_io_codec(f: ta.IO, mode: str) -> ta.ContextManager[ta.IO]:
#     with gzip.open(f, mode) as o:
#         yield o



"""
TODO:
 - Rewrite lol
 - Enum - enforce Abstract or Final
"""
import abc
import collections
import dataclasses as dc
import typing as ta

from ... import lang
from .api import dataclass
from .api import field  # noqa
from .params import MetaclassParams
from .params import get_metaclass_params
from .params import get_params
from .params import get_params_extras


T = ta.TypeVar('T')


def confer_kwarg(out: dict[str, ta.Any], k: str, v: ta.Any) -> None:
    if k in out:
        if out[k] != v:
            raise ValueError
    else:
        out[k] = v


def confer_kwargs(
        bases: ta.Sequence[type],
        kwargs: ta.Mapping[str, ta.Any],
) -> dict[str, ta.Any]:
    out: dict[str, ta.Any] = {}
    for base in bases:
        if not dc.is_dataclass(base):
            continue

        if not (bmp := get_metaclass_params(base)).confer:
            continue

        for ck in bmp.confer:
            if ck in kwargs:
                continue

            if ck in (
                    'frozen',
                    'kw_only',
            ):
                confer_kwarg(out, ck, getattr(get_params(base), ck))

            elif ck in (
                    'cache_hash',
                    'generic_init',
                    'reorder',
            ):
                confer_kwarg(out, ck, getattr(get_params_extras(base), ck))

            elif ck in (
                    'confer',
            ):
                confer_kwarg(out, ck, getattr(bmp, ck))

            else:
                raise KeyError(ck)

    return out


class DataMeta(abc.ABCMeta):
    def __new__(
            mcls,
            name,
            bases,
            namespace,
            *,

            # confer=frozenset(),

            metadata=None,
            **kwargs,
    ):
        cls = lang.super_meta(
            super(),
            mcls,
            name,
            bases,
            namespace,
        )

        ckw = confer_kwargs(bases, kwargs)
        nkw = {**kwargs, **ckw}

        mcp = MetaclassParams(
            confer=nkw.pop('confer', frozenset()),
        )

        mmd = {
            MetaclassParams: mcp,
        }
        if metadata is not None:
            metadata = collections.ChainMap(mmd, metadata)
        else:
            metadata = mmd

        return dataclass(cls, metadata=metadata, **nkw)


# @ta.dataclass_transform(field_specifiers=(field,))  # FIXME: ctor
class Data(
    eq=False,
    order=False,
    metaclass=DataMeta,
):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __post_init__(self, *args, **kwargs) -> None:
        try:
            spi = super().__post_init__  # type: ignore  # noqa
        except AttributeError:
            if args or kwargs:
                raise TypeError(args, kwargs) from None
        else:
            spi(*args, **kwargs)


class Frozen(
    Data,
    frozen=True,
    eq=False,
    order=False,
    confer=frozenset([
        'frozen',
        'cache_hash',
        'confer',
        'reorder',
    ]),
):
    pass


class Box(
    Frozen,
    ta.Generic[T],
    generic_init=True,
    confer=frozenset([
        'frozen',
        'cache_hash',
        'generic_init',
        'confer',
    ]),
):
    v: T



import typing as ta

from ... import lang
from .items import Edge
from .items import Graph
from .items import Node


T = ta.TypeVar('T')


def make_simple(graph: ta.Mapping[T, ta.Iterable[T]]) -> Graph:
    return Graph([
        *[Node(n) for n in {*graph, *lang.flatten(graph.values())}],
        *[Edge(k, v) for k, vs in graph.items() for v in vs],
    ])



"""
TODO:
 - translate json keys
 - debug
"""
# ruff: noqa: UP006 UP007 N802
import contextlib
import datetime
import logging
import threading
import typing as ta

from .json import json_dumps_compact


log = logging.getLogger(__name__)


##


class TidLogFilter(logging.Filter):

    def filter(self, record):
        record.tid = threading.get_native_id()
        return True


##


class JsonLogFormatter(logging.Formatter):

    KEYS: ta.Mapping[str, bool] = {
        'name': False,
        'msg': False,
        'args': False,
        'levelname': False,
        'levelno': False,
        'pathname': False,
        'filename': False,
        'module': False,
        'exc_info': True,
        'exc_text': True,
        'stack_info': True,
        'lineno': False,
        'funcName': False,
        'created': False,
        'msecs': False,
        'relativeCreated': False,
        'thread': False,
        'threadName': False,
        'processName': False,
        'process': False,
    }

    def format(self, record: logging.LogRecord) -> str:
        dct = {
            k: v
            for k, o in self.KEYS.items()
            for v in [getattr(record, k)]
            if not (o and v is None)
        }
        return json_dumps_compact(dct)


##


STANDARD_LOG_FORMAT_PARTS = [
    ('asctime', '%(asctime)-15s'),
    ('process', 'pid=%(process)-6s'),
    ('thread', 'tid=%(thread)x'),
    ('levelname', '%(levelname)s'),
    ('name', '%(name)s'),
    ('separator', '::'),
    ('message', '%(message)s'),
]


class StandardLogFormatter(logging.Formatter):

    @staticmethod
    def build_log_format(parts: ta.Iterable[ta.Tuple[str, str]]) -> str:
        return ' '.join(v for k, v in parts)

    converter = datetime.datetime.fromtimestamp  # type: ignore

    def formatTime(self, record, datefmt=None):
        ct = self.converter(record.created)  # type: ignore
        if datefmt:
            return ct.strftime(datefmt)  # noqa
        else:
            t = ct.strftime('%Y-%m-%d %H:%M:%S')
            return '%s.%03d' % (t, record.msecs)


##


class ProxyLogFilterer(logging.Filterer):
    def __init__(self, underlying: logging.Filterer) -> None:  # noqa
        self._underlying = underlying

    @property
    def underlying(self) -> logging.Filterer:
        return self._underlying

    @property
    def filters(self):
        return self._underlying.filters

    @filters.setter
    def filters(self, filters):
        self._underlying.filters = filters

    def addFilter(self, filter):  # noqa
        self._underlying.addFilter(filter)

    def removeFilter(self, filter):  # noqa
        self._underlying.removeFilter(filter)

    def filter(self, record):
        return self._underlying.filter(record)


class ProxyLogHandler(ProxyLogFilterer, logging.Handler):
    def __init__(self, underlying: logging.Handler) -> None:  # noqa
        ProxyLogFilterer.__init__(self, underlying)

    _underlying: logging.Handler

    @property
    def underlying(self) -> logging.Handler:
        return self._underlying

    def get_name(self):
        return self._underlying.get_name()

    def set_name(self, name):
        self._underlying.set_name(name)

    @property
    def name(self):
        return self._underlying.name

    @property
    def level(self):
        return self._underlying.level

    @level.setter
    def level(self, level):
        self._underlying.level = level

    @property
    def formatter(self):
        return self._underlying.formatter

    @formatter.setter
    def formatter(self, formatter):
        self._underlying.formatter = formatter

    def createLock(self):
        self._underlying.createLock()

    def acquire(self):
        self._underlying.acquire()

    def release(self):
        self._underlying.release()

    def setLevel(self, level):
        self._underlying.setLevel(level)

    def format(self, record):
        return self._underlying.format(record)

    def emit(self, record):
        self._underlying.emit(record)

    def handle(self, record):
        return self._underlying.handle(record)

    def setFormatter(self, fmt):
        self._underlying.setFormatter(fmt)

    def flush(self):
        self._underlying.flush()

    def close(self):
        self._underlying.close()

    def handleError(self, record):
        self._underlying.handleError(record)


##


class StandardLogHandler(ProxyLogHandler):
    pass


##


@contextlib.contextmanager
def _locking_logging_module_lock() -> ta.Iterator[None]:
    if hasattr(logging, '_acquireLock'):
        logging._acquireLock()  # noqa
        try:
            yield
        finally:
            logging._releaseLock()  # type: ignore  # noqa

    elif hasattr(logging, '_lock'):
        # https://github.com/python/cpython/commit/74723e11109a320e628898817ab449b3dad9ee96
        with logging._lock:  # noqa
            yield

    else:
        raise Exception("Can't find lock in logging module")


def configure_standard_logging(
        level: ta.Union[int, str] = logging.INFO,
        *,
        json: bool = False,
        target: ta.Optional[logging.Logger] = None,
        force: bool = False,
        handler_factory: ta.Optional[ta.Callable[[], logging.Handler]] = None,
) -> ta.Optional[StandardLogHandler]:
    with _locking_logging_module_lock():
        if target is None:
            target = logging.root

        #

        if not force:
            if any(isinstance(h, StandardLogHandler) for h in list(target.handlers)):
                return None

        #

        if handler_factory is not None:
            handler = handler_factory()
        else:
            handler = logging.StreamHandler()

        #

        formatter: logging.Formatter
        if json:
            formatter = JsonLogFormatter()
        else:
            formatter = StandardLogFormatter(StandardLogFormatter.build_log_format(STANDARD_LOG_FORMAT_PARTS))
        handler.setFormatter(formatter)

        #

        handler.addFilter(TidLogFilter())

        #

        target.addHandler(handler)

        #

        if level is not None:
            target.setLevel(level)

        #

        return StandardLogHandler(handler)



# type: ignore
# ruff: noqa
# flake8: noqa
#
# Copyright (c) 2012-2017 The ANTLR Project. All rights reserved.
# Use of this file is governed by the BSD 3-clause license that
# can be found in the LICENSE.txt file in the project root.
#


# A set of utility routines useful for all kinds of ANTLR trees.#
from io import StringIO
from ..Token import Token
from ..Utils import escapeWhitespace
from .Tree import RuleNode, ErrorNode, TerminalNode, Tree, ParseTree

# need forward declaration
Parser  = None

class Trees(object):

     # Print out a whole tree in LISP form. {@link #getNodeText} is used on the
    #  node payloads to get the text for the nodes.  Detect
    #  parse trees and extract data appropriately.
    @classmethod
    def toStringTree(cls, t:Tree, ruleNames:list=None, recog:Parser=None):
        if recog is not None:
            ruleNames = recog.ruleNames
        s = escapeWhitespace(cls.getNodeText(t, ruleNames), False)
        if t.getChildCount()==0:
            return s
        with StringIO() as buf:
            buf.write("(")
            buf.write(s)
            buf.write(' ')
            for i in range(0, t.getChildCount()):
                if i > 0:
                    buf.write(' ')
                buf.write(cls.toStringTree(t.getChild(i), ruleNames))
            buf.write(")")
            return buf.getvalue()

    @classmethod
    def getNodeText(cls, t:Tree, ruleNames:list=None, recog:Parser=None):
        if recog is not None:
            ruleNames = recog.ruleNames
        if ruleNames is not None:
            if isinstance(t, RuleNode):
                if t.getAltNumber()!=0: # should use ATN.INVALID_ALT_NUMBER but won't compile
                    return ruleNames[t.getRuleIndex()]+":"+str(t.getAltNumber())
                return ruleNames[t.getRuleIndex()]
            elif isinstance( t, ErrorNode):
                return str(t)
            elif isinstance(t, TerminalNode):
                if t.symbol is not None:
                    return t.symbol.text
        # no recog for rule names
        payload = t.getPayload()
        if isinstance(payload, Token ):
            return payload.text
        return str(t.getPayload())


    # Return ordered list of all children of this node
    @classmethod
    def getChildren(cls, t:Tree):
        return [ t.getChild(i) for i in range(0, t.getChildCount()) ]

    # Return a list of all ancestors of this node.  The first node of
    #  list is the root and the last is the parent of this node.
    #
    @classmethod
    def getAncestors(cls, t:Tree):
        ancestors = []
        t = t.getParent()
        while t is not None:
            ancestors.insert(0, t) # insert at start
            t = t.getParent()
        return ancestors

    @classmethod
    def findAllTokenNodes(cls, t:ParseTree, ttype:int):
        return cls.findAllNodes(t, ttype, True)

    @classmethod
    def findAllRuleNodes(cls, t:ParseTree, ruleIndex:int):
        return cls.findAllNodes(t, ruleIndex, False)

    @classmethod
    def findAllNodes(cls, t:ParseTree, index:int, findTokens:bool):
        nodes = []
        cls._findAllNodes(t, index, findTokens, nodes)
        return nodes

    @classmethod
    def _findAllNodes(cls, t:ParseTree, index:int, findTokens:bool, nodes:list):
        from ..ParserRuleContext import ParserRuleContext
        # check this node (the root) first
        if findTokens and isinstance(t, TerminalNode):
            if t.symbol.type==index:
                nodes.append(t)
        elif not findTokens and isinstance(t, ParserRuleContext):
            if t.ruleIndex == index:
                nodes.append(t)
        # check children
        for i in range(0, t.getChildCount()):
            cls._findAllNodes(t.getChild(i), index, findTokens, nodes)

    @classmethod
    def descendants(cls, t:ParseTree):
        nodes = [t]
        for i in range(0, t.getChildCount()):
            nodes.extend(cls.descendants(t.getChild(i)))
        return nodes



"""
TODO:
 - pickle stdlib objs? have to pin to 3.8 pickle protocol, will be cross-version
 - nonstrict toggle
"""
# ruff: noqa: UP006 UP007
import abc
import base64
import collections.abc
import dataclasses as dc  # noqa
import datetime
import decimal
import enum
import fractions
import functools
import threading
import typing as ta
import uuid
import weakref  # noqa

from .check import check_isinstance
from .check import check_not_none
from .reflect import deep_subclasses
from .reflect import get_optional_alias_arg
from .reflect import is_generic_alias
from .reflect import is_union_alias


T = ta.TypeVar('T')


##


class ObjMarshaler(abc.ABC):
    @abc.abstractmethod
    def marshal(self, o: ta.Any) -> ta.Any:
        raise NotImplementedError

    @abc.abstractmethod
    def unmarshal(self, o: ta.Any) -> ta.Any:
        raise NotImplementedError


class NopObjMarshaler(ObjMarshaler):
    def marshal(self, o: ta.Any) -> ta.Any:
        return o

    def unmarshal(self, o: ta.Any) -> ta.Any:
        return o


@dc.dataclass()
class ProxyObjMarshaler(ObjMarshaler):
    m: ta.Optional[ObjMarshaler] = None

    def marshal(self, o: ta.Any) -> ta.Any:
        return check_not_none(self.m).marshal(o)

    def unmarshal(self, o: ta.Any) -> ta.Any:
        return check_not_none(self.m).unmarshal(o)


@dc.dataclass(frozen=True)
class CastObjMarshaler(ObjMarshaler):
    ty: type

    def marshal(self, o: ta.Any) -> ta.Any:
        return o

    def unmarshal(self, o: ta.Any) -> ta.Any:
        return self.ty(o)


class DynamicObjMarshaler(ObjMarshaler):
    def marshal(self, o: ta.Any) -> ta.Any:
        return marshal_obj(o)

    def unmarshal(self, o: ta.Any) -> ta.Any:
        return o


@dc.dataclass(frozen=True)
class Base64ObjMarshaler(ObjMarshaler):
    ty: type

    def marshal(self, o: ta.Any) -> ta.Any:
        return base64.b64encode(o).decode('ascii')

    def unmarshal(self, o: ta.Any) -> ta.Any:
        return self.ty(base64.b64decode(o))


@dc.dataclass(frozen=True)
class EnumObjMarshaler(ObjMarshaler):
    ty: type

    def marshal(self, o: ta.Any) -> ta.Any:
        return o.name

    def unmarshal(self, o: ta.Any) -> ta.Any:
        return self.ty.__members__[o]  # type: ignore


@dc.dataclass(frozen=True)
class OptionalObjMarshaler(ObjMarshaler):
    item: ObjMarshaler

    def marshal(self, o: ta.Any) -> ta.Any:
        if o is None:
            return None
        return self.item.marshal(o)

    def unmarshal(self, o: ta.Any) -> ta.Any:
        if o is None:
            return None
        return self.item.unmarshal(o)


@dc.dataclass(frozen=True)
class MappingObjMarshaler(ObjMarshaler):
    ty: type
    km: ObjMarshaler
    vm: ObjMarshaler

    def marshal(self, o: ta.Any) -> ta.Any:
        return {self.km.marshal(k): self.vm.marshal(v) for k, v in o.items()}

    def unmarshal(self, o: ta.Any) -> ta.Any:
        return self.ty((self.km.unmarshal(k), self.vm.unmarshal(v)) for k, v in o.items())


@dc.dataclass(frozen=True)
class IterableObjMarshaler(ObjMarshaler):
    ty: type
    item: ObjMarshaler

    def marshal(self, o: ta.Any) -> ta.Any:
        return [self.item.marshal(e) for e in o]

    def unmarshal(self, o: ta.Any) -> ta.Any:
        return self.ty(self.item.unmarshal(e) for e in o)


@dc.dataclass(frozen=True)
class DataclassObjMarshaler(ObjMarshaler):
    ty: type
    fs: ta.Mapping[str, ObjMarshaler]
    nonstrict: bool = False

    def marshal(self, o: ta.Any) -> ta.Any:
        return {k: m.marshal(getattr(o, k)) for k, m in self.fs.items()}

    def unmarshal(self, o: ta.Any) -> ta.Any:
        return self.ty(**{k: self.fs[k].unmarshal(v) for k, v in o.items() if not self.nonstrict or k in self.fs})


@dc.dataclass(frozen=True)
class PolymorphicObjMarshaler(ObjMarshaler):
    class Impl(ta.NamedTuple):
        ty: type
        tag: str
        m: ObjMarshaler

    impls_by_ty: ta.Mapping[type, Impl]
    impls_by_tag: ta.Mapping[str, Impl]

    def marshal(self, o: ta.Any) -> ta.Any:
        impl = self.impls_by_ty[type(o)]
        return {impl.tag: impl.m.marshal(o)}

    def unmarshal(self, o: ta.Any) -> ta.Any:
        [(t, v)] = o.items()
        impl = self.impls_by_tag[t]
        return impl.m.unmarshal(v)


@dc.dataclass(frozen=True)
class DatetimeObjMarshaler(ObjMarshaler):
    ty: type

    def marshal(self, o: ta.Any) -> ta.Any:
        return o.isoformat()

    def unmarshal(self, o: ta.Any) -> ta.Any:
        return self.ty.fromisoformat(o)  # type: ignore


class DecimalObjMarshaler(ObjMarshaler):
    def marshal(self, o: ta.Any) -> ta.Any:
        return str(check_isinstance(o, decimal.Decimal))

    def unmarshal(self, v: ta.Any) -> ta.Any:
        return decimal.Decimal(check_isinstance(v, str))


class FractionObjMarshaler(ObjMarshaler):
    def marshal(self, o: ta.Any) -> ta.Any:
        fr = check_isinstance(o, fractions.Fraction)
        return [fr.numerator, fr.denominator]

    def unmarshal(self, v: ta.Any) -> ta.Any:
        num, denom = check_isinstance(v, list)
        return fractions.Fraction(num, denom)


class UuidObjMarshaler(ObjMarshaler):
    def marshal(self, o: ta.Any) -> ta.Any:
        return str(o)

    def unmarshal(self, o: ta.Any) -> ta.Any:
        return uuid.UUID(o)


##


_DEFAULT_OBJ_MARSHALERS: ta.Dict[ta.Any, ObjMarshaler] = {
    **{t: NopObjMarshaler() for t in (type(None),)},
    **{t: CastObjMarshaler(t) for t in (int, float, str, bool)},
    **{t: Base64ObjMarshaler(t) for t in (bytes, bytearray)},
    **{t: IterableObjMarshaler(t, DynamicObjMarshaler()) for t in (list, tuple, set, frozenset)},
    **{t: MappingObjMarshaler(t, DynamicObjMarshaler(), DynamicObjMarshaler()) for t in (dict,)},

    ta.Any: DynamicObjMarshaler(),

    **{t: DatetimeObjMarshaler(t) for t in (datetime.date, datetime.time, datetime.datetime)},
    decimal.Decimal: DecimalObjMarshaler(),
    fractions.Fraction: FractionObjMarshaler(),
    uuid.UUID: UuidObjMarshaler(),
}

_OBJ_MARSHALER_GENERIC_MAPPING_TYPES: ta.Dict[ta.Any, type] = {
    **{t: t for t in (dict,)},
    **{t: dict for t in (collections.abc.Mapping, collections.abc.MutableMapping)},
}

_OBJ_MARSHALER_GENERIC_ITERABLE_TYPES: ta.Dict[ta.Any, type] = {
    **{t: t for t in (list, tuple, set, frozenset)},
    collections.abc.Set: frozenset,
    collections.abc.MutableSet: set,
    collections.abc.Sequence: tuple,
    collections.abc.MutableSequence: list,
}


def _make_obj_marshaler(
        ty: ta.Any,
        rec: ta.Callable[[ta.Any], ObjMarshaler],
        *,
        nonstrict_dataclasses: bool = False,
) -> ObjMarshaler:
    if isinstance(ty, type):
        if abc.ABC in ty.__bases__:
            impls = [  # type: ignore
                PolymorphicObjMarshaler.Impl(
                    ity,
                    ity.__qualname__,
                    rec(ity),
                )
                for ity in deep_subclasses(ty)
                if abc.ABC not in ity.__bases__
            ]
            return PolymorphicObjMarshaler(
                {i.ty: i for i in impls},
                {i.tag: i for i in impls},
            )

        if issubclass(ty, enum.Enum):
            return EnumObjMarshaler(ty)

        if dc.is_dataclass(ty):
            return DataclassObjMarshaler(
                ty,
                {f.name: rec(f.type) for f in dc.fields(ty)},
                nonstrict=nonstrict_dataclasses,
            )

    if is_generic_alias(ty):
        try:
            mt = _OBJ_MARSHALER_GENERIC_MAPPING_TYPES[ta.get_origin(ty)]
        except KeyError:
            pass
        else:
            k, v = ta.get_args(ty)
            return MappingObjMarshaler(mt, rec(k), rec(v))

        try:
            st = _OBJ_MARSHALER_GENERIC_ITERABLE_TYPES[ta.get_origin(ty)]
        except KeyError:
            pass
        else:
            [e] = ta.get_args(ty)
            return IterableObjMarshaler(st, rec(e))

        if is_union_alias(ty):
            return OptionalObjMarshaler(rec(get_optional_alias_arg(ty)))

    raise TypeError(ty)


##


_OBJ_MARSHALERS_LOCK = threading.RLock()

_OBJ_MARSHALERS: ta.Dict[ta.Any, ObjMarshaler] = dict(_DEFAULT_OBJ_MARSHALERS)

_OBJ_MARSHALER_PROXIES: ta.Dict[ta.Any, ProxyObjMarshaler] = {}


def register_opj_marshaler(ty: ta.Any, m: ObjMarshaler) -> None:
    with _OBJ_MARSHALERS_LOCK:
        if ty in _OBJ_MARSHALERS:
            raise KeyError(ty)
        _OBJ_MARSHALERS[ty] = m


def get_obj_marshaler(
        ty: ta.Any,
        *,
        no_cache: bool = False,
        **kwargs: ta.Any,
) -> ObjMarshaler:
    with _OBJ_MARSHALERS_LOCK:
        if not no_cache:
            try:
                return _OBJ_MARSHALERS[ty]
            except KeyError:
                pass

        try:
            return _OBJ_MARSHALER_PROXIES[ty]
        except KeyError:
            pass

        rec = functools.partial(
            get_obj_marshaler,
            no_cache=no_cache,
            **kwargs,
        )

        p = ProxyObjMarshaler()
        _OBJ_MARSHALER_PROXIES[ty] = p
        try:
            m = _make_obj_marshaler(ty, rec, **kwargs)
        finally:
            del _OBJ_MARSHALER_PROXIES[ty]
        p.m = m

        if not no_cache:
            _OBJ_MARSHALERS[ty] = m
        return m


##


def marshal_obj(o: ta.Any, ty: ta.Any = None) -> ta.Any:
    return get_obj_marshaler(ty if ty is not None else type(o)).marshal(o)


def unmarshal_obj(o: ta.Any, ty: ta.Union[ta.Type[T], ta.Any]) -> T:
    return get_obj_marshaler(ty).unmarshal(o)



"""
TODO:
 - proper ManagedProvider[Impl], DeferringProvider[Impl], closing (-> Deferring('close'))
"""
import contextlib
import typing as ta

from .. import lang
from .binder import bind
from .elements import Elemental
from .impl.inspect import build_kwargs_target
from .injector import Injector
from .injector import create_injector


if ta.TYPE_CHECKING:
    from .. import asyncs as _asyncs
else:
    _asyncs = lang.proxy_import('..asyncs', __package__)


T = ta.TypeVar('T')


##


@contextlib.contextmanager
def create_managed_injector(*args: Elemental) -> ta.Generator[Injector, None, None]:
    i = create_injector(
        bind(contextlib.ExitStack, singleton=True, eager=True),
        *args,
    )
    with i[contextlib.ExitStack]:
        yield i


def make_managed_provider(
        fac: ta.Callable[..., T],
        *fns: ta.Callable[[T], ta.ContextManager[T]],
) -> ta.Callable[..., T]:
    kt = build_kwargs_target(fac)

    def _provide(
            i: Injector,
            es: contextlib.ExitStack,
    ):
        obj = i.inject(kt)
        if not fns:
            obj = es.enter_context(obj)
        else:
            for fn in fns:
                es.enter_context(fn(obj))
        return obj

    return _provide


##


@contextlib.asynccontextmanager
async def create_async_managed_injector(*args: Elemental) -> ta.AsyncGenerator[Injector, None]:
    i = await _asyncs.s_to_a(create_injector)(
        bind(contextlib.AsyncExitStack, singleton=True, eager=True),
        *args,
    )
    async with i[contextlib.AsyncExitStack]:
        yield i


def make_async_managed_provider(
        fac: ta.Callable[..., T],
        *fns: ta.Callable[[T], ta.AsyncContextManager[T]],
) -> ta.Callable[..., T]:
    kt = build_kwargs_target(fac)

    def _provide(
            i: Injector,
            aes: contextlib.AsyncExitStack,
    ):
        obj = i.inject(kt)
        if not fns:
            obj = _asyncs.a_to_s(aes.enter_async_context)(obj)
        else:
            for fn in fns:
                _asyncs.a_to_s(aes.enter_async_context)(fn(obj))
        return obj

    return _provide



import numbers
import operator
import typing as ta

from ... import check
from ... import lang
from .ast import AndExpression
from .ast import Arithmetic
from .ast import ArithmeticOperator
from .ast import ArithmeticUnary
from .ast import Assign
from .ast import Comparator
from .ast import ComparatorName
from .ast import CurrentNode
from .ast import Expref
from .ast import Field
from .ast import FilterProjection
from .ast import Flatten
from .ast import FunctionExpression
from .ast import Identity
from .ast import Index
from .ast import IndexExpression
from .ast import KeyValPair
from .ast import LetExpression
from .ast import Literal
from .ast import MultiSelectDict
from .ast import MultiSelectList
from .ast import Node
from .ast import NotExpression
from .ast import OrExpression
from .ast import Pipe
from .ast import Projection
from .ast import RootNode
from .ast import Slice
from .ast import Subexpression
from .ast import UnaryArithmeticOperator
from .ast import ValueProjection
from .ast import VariableRef
from .exceptions import UndefinedVariableError
from .functions import DefaultFunctions
from .functions import Functions
from .scope import ScopedChainDict


if ta.TYPE_CHECKING:
    import html
    import json
else:
    html = lang.proxy_import('html')
    json = lang.proxy_import('json')


##


def _equals(x: ta.Any, y: ta.Any) -> bool:
    if _is_special_number_case(x, y):
        return False
    else:
        return x == y


def _is_special_number_case(x: ta.Any, y: ta.Any) -> bool | None:
    # We need to special case comparing 0 or 1 to True/False.  While normally comparing any integer other than 0/1 to
    # True/False will always return False.  However 0/1 have this:
    # >>> 0 == True
    # False
    # >>> 0 == False
    # True
    # >>> 1 == True
    # True
    # >>> 1 == False
    # False
    #
    # Also need to consider that:
    # >>> 0 in [True, False]
    # True
    if _is_actual_number(x) and x in (0, 1):
        return isinstance(y, bool)

    elif _is_actual_number(y) and y in (0, 1):
        return isinstance(x, bool)

    else:
        return None


def _is_comparable(x: ta.Any) -> bool:
    # The spec doesn't officially support string types yet, but enough people are relying on this behavior that it's
    # been added back.  This should eventually become part of the official spec.
    return _is_actual_number(x) or isinstance(x, str)


def _is_actual_number(x: ta.Any) -> bool:
    # We need to handle python's quirkiness with booleans, specifically:
    #
    # >>> isinstance(False, int)
    # True
    # >>> isinstance(True, int)
    # True
    if isinstance(x, bool):
        return False
    return isinstance(x, numbers.Number)


def node_type(n: Node) -> str:
    return lang.snake_case(type(n).__name__)


##


class Visitor:
    def __init__(self) -> None:
        super().__init__()

        self._method_cache: dict[str, ta.Callable] = {}

    def visit(self, node: Node, *args: ta.Any, **kwargs: ta.Any) -> ta.Any:
        nty = node_type(node)
        method = self._method_cache.get(nty)
        if method is None:
            method = check.not_none(getattr(self, f'visit_{nty}', self.default_visit))
            self._method_cache[nty] = method
        return method(node, *args, **kwargs)

    def default_visit(self, node, *args, **kwargs):
        raise NotImplementedError('default_visit')


##


class Options:
    """Options to control how a Jmespath function is evaluated."""

    def __init__(
            self,
            dict_cls: type | None = None,
            custom_functions: Functions | None = None,
            enable_legacy_literals: bool = False,
    ) -> None:
        super().__init__()

        # The class to use when creating a dict.  The interpreter may create dictionaries during the evaluation of a
        # Jmespath expression.  For example, a multi-select hash will create a dictionary.  By default we use a dict()
        # type. You can set this value to change what dict type is used. The most common reason you would change this is
        # if you want to set a collections.OrderedDict so that you can have predictable key ordering.
        self.dict_cls = dict_cls
        self.custom_functions = custom_functions

        # The flag to enable pre-JEP-12 literal compatibility.
        # JEP-12 deprecates `foo` -> "foo" syntax.
        # Valid expressions MUST use: `"foo"` -> "foo"
        # Setting this flag to `True` enables support for legacy syntax.
        self.enable_legacy_literals = enable_legacy_literals


class _Expression:
    def __init__(self, expression: Node, interpreter: 'TreeInterpreter') -> None:
        super().__init__()
        self.expression = expression
        self.interpreter = interpreter

    def visit(self, node: Node, *args: ta.Any, **kwargs: ta.Any) -> ta.Any:
        return self.interpreter.visit(node, *args, **kwargs)


class TreeInterpreter(Visitor):
    _COMPARATOR_FUNC: ta.Mapping[ComparatorName, ta.Callable] = {
        'eq': _equals,
        'ne': lambda x, y: not _equals(x, y),
        'lt': operator.lt,
        'gt': operator.gt,
        'lte': operator.le,
        'gte': operator.ge,
    }

    _EQUALITY_OPS: ta.AbstractSet[ComparatorName] = {
        'eq',
        'ne',
    }

    _ARITHMETIC_FUNC: ta.Mapping[ArithmeticOperator, ta.Callable] = {
        'div': operator.floordiv,
        'divide': operator.truediv,
        'minus': operator.sub,
        'modulo': operator.mod,
        'multiply': operator.mul,
        'plus': operator.add,
    }

    _ARITHMETIC_UNARY_FUNC: ta.Mapping[UnaryArithmeticOperator, ta.Callable] = {
        'minus': operator.neg,
        'plus': lambda x: x,
    }

    def __init__(self, options: Options | None = None) -> None:
        super().__init__()

        self._dict_cls: type = dict

        if options is None:
            options = Options()
        self._options = options

        if options.dict_cls is not None:
            self._dict_cls = options.dict_cls

        if options.custom_functions is not None:
            self._functions: Functions = options.custom_functions
        else:
            self._functions = DefaultFunctions()

        self._root: Node | None = None
        self._scope: ScopedChainDict = ScopedChainDict()

    def default_visit(self, node: Node, *args: ta.Any, **kwargs: ta.Any) -> ta.NoReturn:
        raise NotImplementedError(node_type(node))

    def evaluate(self, ast: Node, root: Node) -> ta.Any:
        self._root = root
        return self.visit(ast, root)

    def visit_subexpression(self, node: Subexpression, value: ta.Any) -> ta.Any:
        result = value
        for child in node.children_nodes:
            result = self.visit(child, result)
            if result is None:
                return None
        return result

    def visit_field(self, node: Field, value: ta.Any) -> ta.Any:
        try:
            return value.get(node.name)
        except AttributeError:
            return None

    def visit_comparator(self, node: Comparator, value: ta.Any) -> ta.Any:
        # Common case: comparator is == or !=
        comparator_func = self._COMPARATOR_FUNC[node.name]
        if node.name in self._EQUALITY_OPS:
            return comparator_func(
                self.visit(node.first, value),
                self.visit(node.second, value),
            )

        else:
            # Ordering operators are only valid for numbers. Evaluating any other type with a comparison operator will
            # yield a None value.
            left = self.visit(node.first, value)
            right = self.visit(node.second, value)
            # num_types = (int, float)
            if not (_is_comparable(left) and _is_comparable(right)):
                return None
            return comparator_func(left, right)

    def visit_arithmetic_unary(self, node: ArithmeticUnary, value: ta.Any) -> ta.Any:
        operation = self._ARITHMETIC_UNARY_FUNC[node.operator]
        return operation(
            self.visit(node.expression, value),
        )

    def visit_arithmetic(self, node: Arithmetic, value: ta.Any) -> ta.Any:
        operation = self._ARITHMETIC_FUNC[node.operator]
        return operation(
            self.visit(node.left, value),
            self.visit(node.right, value),
        )

    def visit_current_node(self, node: CurrentNode, value: ta.Any) -> ta.Any:
        return value

    def visit_root_node(self, node: RootNode, value: ta.Any) -> ta.Any:
        return self._root

    def visit_expref(self, node: Expref, value: ta.Any) -> ta.Any:
        return _Expression(node.expression, self)

    def visit_function_expression(self, node: FunctionExpression, value: ta.Any) -> ta.Any:
        resolved_args = []
        for child in node.args:
            current = self.visit(child, value)
            resolved_args.append(current)

        return self._functions.call_function(node.name, resolved_args)

    def visit_filter_projection(self, node: FilterProjection, value: ta.Any) -> ta.Any:
        base = self.visit(node.left, value)
        if not isinstance(base, list):
            return None

        comparator_node = node.comparator
        collected = []
        for element in base:
            if self._is_true(self.visit(comparator_node, element)):
                current = self.visit(node.right, element)
                if current is not None:
                    collected.append(current)

        return collected

    def visit_flatten(self, node: Flatten, value: ta.Any) -> ta.Any:
        base = self.visit(node.node, value)
        if not isinstance(base, list):
            # Can't flatten the object if it's not a list.
            return None

        merged_list = []
        for element in base:
            if isinstance(element, list):
                merged_list.extend(element)
            else:
                merged_list.append(element)

        return merged_list

    def visit_identity(self, node: Identity, value: ta.Any) -> ta.Any:
        return value

    def visit_index(self, node: Index, value: ta.Any) -> ta.Any:
        # Even though we can index strings, we don't want to support that.
        if not isinstance(value, list):
            return None

        try:
            return value[node.index]
        except IndexError:
            return None

    def visit_index_expression(self, node: IndexExpression, value: ta.Any) -> ta.Any:
        result = value
        for child in node.nodes:
            result = self.visit(child, result)

        return result

    def visit_slice(self, node: Slice, value: ta.Any) -> ta.Any:
        if isinstance(value, str):
            return value[node.start:node.end:node.step]

        if not isinstance(value, list):
            return None

        s = slice(node.start, node.end, node.step)
        return value[s]

    def visit_key_val_pair(self, node: KeyValPair, value: ta.Any) -> ta.Any:
        return self.visit(node.node, value)

    def visit_literal(self, node: Literal, value: ta.Any):
        return node.literal_value

    def visit_multi_select_dict(self, node: MultiSelectDict, value: ta.Any) -> ta.Any:
        collected = self._dict_cls()
        for child in node.nodes:
            collected[child.key_name] = self.visit(child, value)

        return collected

    def visit_multi_select_list(self, node: MultiSelectList, value: ta.Any) -> ta.Any:
        collected = []
        for child in node.nodes:
            collected.append(self.visit(child, value))

        return collected

    def visit_or_expression(self, node: OrExpression, value: ta.Any) -> ta.Any:
        matched = self.visit(node.left, value)

        if self._is_false(matched):
            matched = self.visit(node.right, value)

        return matched

    def visit_and_expression(self, node: AndExpression, value: ta.Any) -> ta.Any:
        matched = self.visit(node.left, value)

        if self._is_false(matched):
            return matched

        return self.visit(node.right, value)

    def visit_not_expression(self, node: NotExpression, value: ta.Any) -> ta.Any:
        original_result = self.visit(node.expr, value)

        if _is_actual_number(original_result) and original_result == 0:
            # Special case for 0, !0 should be false, not true. 0 is not a special cased integer in jmespath.
            return False

        return not original_result

    def visit_pipe(self, node: Pipe, value: ta.Any) -> ta.Any:
        result = value
        for child in [node.left, node.right]:
            result = self.visit(child, result)
        return result

    def visit_projection(self, node: Projection, value: ta.Any) -> ta.Any:
        base = self.visit(node.left, value)

        allow_string = False
        first_child = node.left
        if isinstance(first_child, IndexExpression):
            nested_children = first_child.nodes
            if len(nested_children) > 1 and isinstance(nested_children[1], Slice):
                allow_string = True

        if isinstance(base, str) and allow_string:
            # projections are really sub-expressions in disguise evaluate the rhs when lhs is a sliced string
            return self.visit(node.right, base)

        if not isinstance(base, list):
            return None

        collected = []
        for element in base:
            current = self.visit(node.right, element)
            if current is not None:
                collected.append(current)

        return collected

    def visit_let_expression(self, node: LetExpression, value: ta.Any) -> ta.Any:
        scope = {}
        for assign in node.bindings:
            scope.update(self.visit(assign, value))
        self._scope.push_scope(scope)
        result = self.visit(node.expr, value)
        self._scope.pop_scope()
        return result

    def visit_assign(self, node: Assign, value: ta.Any) -> ta.Any:
        name = node.name
        value = self.visit(node.expr, value)
        return {name: value}

    def visit_variable_ref(self, node: VariableRef, value: ta.Any) -> ta.Any:
        try:
            return self._scope[node.name]
        except KeyError:
            raise UndefinedVariableError(node.name)  # noqa

    def visit_value_projection(self, node: ValueProjection, value: ta.Any) -> ta.Any:
        base = self.visit(node.left, value)
        try:
            base = base.values()
        except AttributeError:
            return None

        collected = []
        for element in base:
            current = self.visit(node.right, element)
            if current is not None:
                collected.append(current)

        return collected

    def _is_false(self, value: ta.Any) -> bool:
        # This looks weird, but we're explicitly using equality checks because the truth/false values are different
        # between python and jmespath.
        return (
            value == '' or  # noqa
            value == [] or
            value == {} or
            value is None or
            value is False
        )

    def _is_true(self, value: ta.Any) -> bool:
        return not self._is_false(value)


class GraphvizVisitor:
    def __init__(self) -> None:
        super().__init__()
        self._lines: list[str] = []
        self._count = 1

    def visit(self, node: Node) -> str:
        self._lines.append('digraph AST {')
        current = f'{node_type(node)}{self._count}'
        self._count += 1
        self._visit(node, current)
        self._lines.append('}')
        return '\n'.join(self._lines)

    def _node_value(self, node: Node) -> lang.Maybe[ta.Any]:
        if isinstance(node, (ArithmeticUnary, Arithmetic)):
            return lang.just(node.operator)
        elif isinstance(node, (Assign, Comparator, FunctionExpression, Field, VariableRef)):
            return lang.just(node.name)
        elif isinstance(node, Index):
            return lang.just(node.index)
        elif isinstance(node, KeyValPair):
            return lang.just(node.key_name)
        elif isinstance(node, Literal):
            return lang.just(node.literal_value)
        else:
            return lang.empty()

    def _visit(self, node: Node, current: str) -> None:
        self._lines.append(
            f'{current} '
            f'[label="{node_type(node)}({self._node_value(node).map(json.dumps).map(html.escape).or_else("")})"]',
        )
        for child in node.children:
            child_name = f'{node_type(child)}{self._count}'
            self._count += 1
            self._lines.append(f'  {current} -> {child_name}')
            self._visit(child, child_name)



import dataclasses as dc
import typing as ta


T = ta.TypeVar('T')


@dc.dataclass(frozen=True)
class Func(ta.Generic[T]):
    fn: ta.Callable[..., T]

    def __call__(self, *args: ta.Any, **kwargs: ta.Any) -> T:
        return self.fn(*args, **kwargs)
